# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import Any, Callable, Container, Dict, Iterable, List

try:
    from systemd.daemon import notify
except ImportError:
    notify = None

from swh.core.statsd import statsd
from swh.model.model import (
    BaseContent,
    BaseModel,
    Content,
    Directory,
    ExtID,
    MetadataAuthority,
    MetadataFetcher,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    RawExtrinsicMetadata,
    Release,
    Revision,
    SkippedContent,
    Snapshot,
)
from swh.storage.exc import HashCollision
from swh.storage.fixer import fix_objects

logger = logging.getLogger(__name__)

GRAPH_OPERATIONS_METRIC = "swh_graph_replayer_operations_total"
GRAPH_DURATION_METRIC = "swh_graph_replayer_duration_seconds"


object_converter_fn: Dict[str, Callable[[Dict], BaseModel]] = {
    "origin": Origin.from_dict,
    "origin_visit": OriginVisit.from_dict,
    "origin_visit_status": OriginVisitStatus.from_dict,
    "snapshot": Snapshot.from_dict,
    "revision": Revision.from_dict,
    "release": Release.from_dict,
    "directory": Directory.from_dict,
    "content": Content.from_dict,
    "skipped_content": SkippedContent.from_dict,
    "metadata_authority": MetadataAuthority.from_dict,
    "metadata_fetcher": MetadataFetcher.from_dict,
    "raw_extrinsic_metadata": RawExtrinsicMetadata.from_dict,
    "extid": ExtID.from_dict,
}


def process_replay_objects(all_objects, *, storage):
    for (object_type, objects) in all_objects.items():
        logger.debug("Inserting %s %s objects", len(objects), object_type)
        with statsd.timed(GRAPH_DURATION_METRIC, tags={"object_type": object_type}):
            _insert_objects(object_type, objects, storage)
        statsd.increment(
            GRAPH_OPERATIONS_METRIC, len(objects), tags={"object_type": object_type}
        )
    if notify:
        notify("WATCHDOG=1")


def collision_aware_content_add(
    content_add_fn: Callable[[Iterable[Any]], None], contents: List[BaseContent]
) -> None:
    """Add contents to storage. If a hash collision is detected, an error is
       logged. Then this adds the other non colliding contents to the storage.

    Args:
        content_add_fn: Storage content callable
        contents: List of contents or skipped contents to add to storage

    """
    if not contents:
        return
    colliding_content_hashes: List[Dict[str, Any]] = []
    while True:
        try:
            content_add_fn(contents)
        except HashCollision as e:
            colliding_content_hashes.append(
                {
                    "algo": e.algo,
                    "hash": e.hash_id,  # hex hash id
                    "objects": e.colliding_contents,  # hex hashes
                }
            )
            colliding_hashes = e.colliding_content_hashes()
            # Drop the colliding contents from the transaction
            contents = [c for c in contents if c.hashes() not in colliding_hashes]
        else:
            # Successfully added contents, we are done
            break
    if colliding_content_hashes:
        for collision in colliding_content_hashes:
            logger.error("Collision detected: %(collision)s", {"collision": collision})


def dict_key_dropper(d: Dict, keys_to_drop: Container) -> Dict:
    """Returns a copy of the dict d without any key listed in keys_to_drop"""
    return {k: v for (k, v) in d.items() if k not in keys_to_drop}


def _insert_objects(object_type: str, objects: List[Dict], storage) -> None:
    """Insert objects of type object_type in the storage.

    """
    objects = fix_objects(object_type, objects)

    if object_type == "content":
        # for bw compat, skipped content should now be delivered in the skipped_content
        # topic
        contents: List[BaseContent] = []
        skipped_contents: List[BaseContent] = []
        for content in objects:
            c = BaseContent.from_dict(content)
            if isinstance(c, SkippedContent):
                logger.warning(
                    "Received a series of skipped_content in the "
                    "content topic, this should not happen anymore"
                )
                skipped_contents.append(c)
            else:
                contents.append(c)
        collision_aware_content_add(storage.skipped_content_add, skipped_contents)
        collision_aware_content_add(storage.content_add_metadata, contents)
    elif object_type == "skipped_content":
        skipped_contents = [SkippedContent.from_dict(obj) for obj in objects]
        collision_aware_content_add(storage.skipped_content_add, skipped_contents)
    elif object_type in ("origin_visit", "origin_visit_status"):
        origins: List[Origin] = []
        converter_fn = object_converter_fn[object_type]
        model_objs = []
        for obj in objects:
            origins.append(Origin(url=obj["origin"]))
            model_objs.append(converter_fn(obj))
        storage.origin_add(origins)
        method = getattr(storage, f"{object_type}_add")
        method(model_objs)
    elif object_type == "raw_extrinsic_metadata":
        converted = [RawExtrinsicMetadata.from_dict(o) for o in objects]
        authorities = {emd.authority for emd in converted}
        fetchers = {emd.fetcher for emd in converted}
        storage.metadata_authority_add(authorities)
        storage.metadata_fetcher_add(fetchers)
        storage.raw_extrinsic_metadata_add(converted)
    elif object_type == "revision":
        # drop the metadata field from the revision (is any); this field is
        # about to be dropped from the data model (in favor of
        # raw_extrinsic_metadata) and there can be bogus values in the existing
        # journal (metadata with \0000 in it)
        method = getattr(storage, object_type + "_add")
        method(
            [
                object_converter_fn[object_type](dict_key_dropper(o, ("metadata",)))
                for o in objects
            ]
        )
    elif object_type in (
        "directory",
        "extid",
        "revision",
        "release",
        "snapshot",
        "origin",
        "metadata_fetcher",
        "metadata_authority",
    ):
        method = getattr(storage, object_type + "_add")
        method([object_converter_fn[object_type](o) for o in objects])
    else:
        logger.warning("Received a series of %s, this should not happen", object_type)



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
from unittest.mock import patch

from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.storage.algos.revisions_walker import get_revisions_walker

# For those tests, we will walk the following revisions history
# with different orderings:
#
# *   commit b364f53155044e5308a0f73abb3b5f01995a5b7d
# |\  Merge: 836d498 b94886c
# | | Author:     Adam <adam.janicki@roche.com>
# | | AuthorDate: Fri Oct 4 12:50:49 2013 +0200
# | | Commit:     Adam <adam.janicki@roche.com>
# | | CommitDate: Fri Oct 4 12:50:49 2013 +0200
# | |
# | |     Merge branch 'release/1.0'
# | |
# | * commit b94886c500c46e32dc3d7ebae8a5409accd592e5
# | | Author:     Adam <adam.janicki@roche.com>
# | | AuthorDate: Fri Oct 4 12:50:38 2013 +0200
# | | Commit:     Adam <adam.janicki@roche.com>
# | | CommitDate: Fri Oct 4 12:50:38 2013 +0200
# | |
# | |     updating poms for 1.0 release
# | |
# | * commit 0cb6b4611d65bee0f57821dac7f611e2f8a02433
# | | Author:     Adam <adam.janicki@roche.com>
# | | AuthorDate: Fri Oct 4 12:50:38 2013 +0200
# | | Commit:     Adam <adam.janicki@roche.com>
# | | CommitDate: Fri Oct 4 12:50:38 2013 +0200
# | |
# | |     updating poms for 1.0 release
# | |
# | * commit 2b0240c6d682bad51532eec15b8a7ed6b75c8d31
# | | Author:     Adam Janicki <janickia>
# | | AuthorDate: Fri Oct 4 12:50:22 2013 +0200
# | | Commit:     Adam Janicki <janickia>
# | | CommitDate: Fri Oct 4 12:50:22 2013 +0200
# | |
# | |     For 1.0 release. Allow untracked.
# | |
# | * commit b401c50863475db4440c85c10ac0b6423b61554d
# | | Author:     Adam <adam.janicki@roche.com>
# | | AuthorDate: Fri Oct 4 12:48:12 2013 +0200
# | | Commit:     Adam <adam.janicki@roche.com>
# | | CommitDate: Fri Oct 4 12:48:12 2013 +0200
# | |
# | |     updating poms for 1.0 release
# | |
# | * commit 9c5051397e5c2e0c258bb639c3dd34406584ca10
# |/  Author:     Adam Janicki <janickia>
# |   AuthorDate: Fri Oct 4 12:47:48 2013 +0200
# |   Commit:     Adam Janicki <janickia>
# |   CommitDate: Fri Oct 4 12:47:48 2013 +0200
# |
# |       For 1.0 release.
# |
# * commit 836d498396fb9b5d45c896885f84d8d60a5651dc
# | Author:     Adam Janicki <janickia>
# | AuthorDate: Fri Oct 4 12:08:16 2013 +0200
# | Commit:     Adam Janicki <janickia>
# | CommitDate: Fri Oct 4 12:08:16 2013 +0200
# |
# |     Add ignores
# |
# * commit ee96c2a2d397b79070d2b6fe3051290963748358
# | Author:     Adam <adam.janicki@roche.com>
# | AuthorDate: Fri Oct 4 10:48:16 2013 +0100
# | Commit:     Adam <adam.janicki@roche.com>
# | CommitDate: Fri Oct 4 10:48:16 2013 +0100
# |
# |     Reset author
# |
# * commit 8f89dda8e072383cf50d42532ae8f52ad89f8fdf
#   Author:     Adam <adam.janicki@roche.com>
#   AuthorDate: Fri Oct 4 02:20:19 2013 -0700
#   Commit:     Adam <adam.janicki@roche.com>
#   CommitDate: Fri Oct 4 02:20:19 2013 -0700
#
#       Initial commit

# raw dump of the above history in swh format
_revisions_list = [
    {
        "author": {
            "email": b"adam.janicki@roche.com",  # noqa
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883849},
        },
        "date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883849},
        },
        "directory": b"\xefX\xe7\xa6\\\xda\xdf\xfdH\xdbH\xfbq\x96@{\x98?9\xfe",
        "id": b"\xb3d\xf51U\x04NS\x08\xa0\xf7:\xbb;_\x01\x99Z[}",
        "message": b"Merge branch 'release/1.0'",
        "metadata": None,
        "parents": [
            b"\x83mI\x83\x96\xfb\x9b]E\xc8\x96\x88_\x84\xd8\xd6\nVQ\xdc",
            b"\xb9H\x86\xc5\x00\xc4n2\xdc=~\xba\xe8\xa5@\x9a\xcc\xd5\x92\xe5",
        ],  # noqa
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883838},
        },
        "date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883838},
        },
        "directory": b"\xefX\xe7\xa6\\\xda\xdf\xfdH\xdbH\xfbq\x96@{\x98?9\xfe",
        "id": b"\xb9H\x86\xc5\x00\xc4n2\xdc=~\xba\xe8\xa5@\x9a\xcc\xd5\x92\xe5",
        "message": b"updating poms for 1.0 release",
        "metadata": None,
        "parents": [
            b"\x0c\xb6\xb4a\x1de\xbe\xe0\xf5x!\xda\xc7\xf6\x11\xe2\xf8\xa0$3"
        ],  # noqa
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883838},
        },
        "date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883838},
        },
        "directory": b"\xefX\xe7\xa6\\\xda\xdf\xfdH\xdbH\xfbq\x96@{\x98?9\xfe",
        "id": b"\x0c\xb6\xb4a\x1de\xbe\xe0\xf5x!\xda\xc7\xf6\x11\xe2\xf8\xa0$3",
        "message": b"updating poms for 1.0 release",
        "metadata": None,
        "parents": [b"+\x02@\xc6\xd6\x82\xba\xd5\x152\xee\xc1[\x8a~\xd6\xb7\\\x8d1"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "id": 8040906,
            "name": b"Adam Janicki",
        },
        "committer": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "id": 8040906,
            "name": b"Adam Janicki",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883822},
        },
        "date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883822},
        },
        "directory": b"\xefX\xe7\xa6\\\xda\xdf\xfdH\xdbH\xfbq\x96@{\x98?9\xfe",
        "id": b"+\x02@\xc6\xd6\x82\xba\xd5\x152\xee\xc1[\x8a~\xd6\xb7\\\x8d1",
        "message": b"For 1.0 release. Allow untracked.\n",
        "metadata": None,
        "parents": [b"\xb4\x01\xc5\x08cG]\xb4D\x0c\x85\xc1\n\xc0\xb6B;aUM"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883692},
        },
        "date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883692},
        },
        "directory": b"d@\xe7\x143w\xcb\xf7\xad\xae\x91\xd5\xec\xd8\x95\x82"
        b"\x02\xa6=\x1b",
        "id": b"\xb4\x01\xc5\x08cG]\xb4D\x0c\x85\xc1\n\xc0\xb6B;aUM",
        "message": b"updating poms for 1.0 release",
        "metadata": None,
        "parents": [b"\x9cPQ9~\\.\x0c%\x8b\xb69\xc3\xdd4@e\x84\xca\x10"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "id": 8040906,
            "name": b"Adam Janicki",
        },
        "committer": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "id": 8040906,
            "name": b"Adam Janicki",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883668},
        },
        "date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380883668},
        },
        "directory": b"\n\x857\x94r\xbe\xcc\x04=\xe9}\xe5\xfd\xdf?nR\xe6\xa7\x9e",
        "id": b"\x9cPQ9~\\.\x0c%\x8b\xb69\xc3\xdd4@e\x84\xca\x10",
        "message": b"For 1.0 release.\n",
        "metadata": None,
        "parents": [b"\x83mI\x83\x96\xfb\x9b]E\xc8\x96\x88_\x84\xd8\xd6\nVQ\xdc"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "id": 8040906,
            "name": b"Adam Janicki",
        },
        "committer": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "id": 8040906,
            "name": b"Adam Janicki",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380881296},
        },
        "date": {
            "negative_utc": None,
            "offset": 120,
            "timestamp": {"microseconds": 0, "seconds": 1380881296},
        },
        "directory": b".\xf9\xa5\xcb\xb0\xd3\xdc\x9b{\xb8\x81\x03l\xe2P\x16c\x0b|\xe6",  # noqa
        "id": b"\x83mI\x83\x96\xfb\x9b]E\xc8\x96\x88_\x84\xd8\xd6\nVQ\xdc",
        "message": b"Add ignores\n",
        "metadata": None,
        "parents": [b"\xee\x96\xc2\xa2\xd3\x97\xb7\x90p\xd2\xb6\xfe0Q)\tct\x83X"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": 60,
            "timestamp": {"microseconds": 0, "seconds": 1380880096},
        },
        "date": {
            "negative_utc": None,
            "offset": 60,
            "timestamp": {"microseconds": 0, "seconds": 1380880096},
        },
        "directory": b"\xc7r\xc4\x9f\xc0$\xd4\xab\xff\xcb]\xf6<\xcb\x8b~\xec\xc4\xd1)",  # noqa
        "id": b"\xee\x96\xc2\xa2\xd3\x97\xb7\x90p\xd2\xb6\xfe0Q)\tct\x83X",
        "message": b"Reset author\n",
        "metadata": None,
        "parents": [b"\x8f\x89\xdd\xa8\xe0r8<\xf5\rBS*\xe8\xf5*\xd8\x9f\x8f\xdf"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "id": 8040905,
            "name": b"Adam",
        },
        "committer_date": {
            "negative_utc": None,
            "offset": -420,
            "timestamp": {"microseconds": 0, "seconds": 1380878419},
        },
        "date": {
            "negative_utc": None,
            "offset": -420,
            "timestamp": {"microseconds": 0, "seconds": 1380878419},
        },
        "directory": b"WS\xbaX\xd6x{q\x8f\x020i\xc5\x95\xa01\xf7y\xb2\x80",
        "id": b"\x8f\x89\xdd\xa8\xe0r8<\xf5\rBS*\xe8\xf5*\xd8\x9f\x8f\xdf",
        "message": b"Initial commit\n",
        "metadata": None,
        "parents": [],
        "synthetic": False,
        "type": "git",
    },
]


_rev_start = "b364f53155044e5308a0f73abb3b5f01995a5b7d"

_rev_missing = "836d498396fb9b5d45c896885f84d8d60a5651dc"


class RevisionsWalkerTest(unittest.TestCase):
    def check_revisions_ordering(
        self, rev_walker_type, expected_result, truncated_history
    ):
        with patch("swh.storage.postgresql.storage.Storage") as MockStorage:
            storage = MockStorage()
            if not truncated_history:
                storage.revision_log.return_value = _revisions_list
            else:
                revs_lists_truncated = [
                    None if hash_to_hex(rev["id"]) == _rev_missing else rev
                    for rev in _revisions_list
                ]

                storage.revision_log.return_value = revs_lists_truncated

            revs_walker = get_revisions_walker(
                rev_walker_type, storage, hash_to_bytes(_rev_start)
            )

            self.assertEqual(
                list(map(hash_to_bytes, expected_result)),
                [rev["id"] for rev in revs_walker],
            )

            self.assertEqual(revs_walker.is_history_truncated(), truncated_history)

            if truncated_history:
                missing_revs = revs_walker.missing_revisions()
                self.assertEqual(missing_revs, {hash_to_bytes(_rev_missing)})
            else:
                self.assertEqual(revs_walker.missing_revisions(), set())

    def test_revisions_walker_committer_date(self):

        # revisions should be returned in reverse chronological order
        # of their committer date
        expected_result = [
            "b364f53155044e5308a0f73abb3b5f01995a5b7d",
            "b94886c500c46e32dc3d7ebae8a5409accd592e5",
            "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
            "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
            "b401c50863475db4440c85c10ac0b6423b61554d",
            "9c5051397e5c2e0c258bb639c3dd34406584ca10",
            "836d498396fb9b5d45c896885f84d8d60a5651dc",
            "ee96c2a2d397b79070d2b6fe3051290963748358",
            "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
        ]

        self.check_revisions_ordering(
            "committer_date", expected_result, truncated_history=False
        )

    def test_revisions_walker_dfs(self):

        # revisions should be returned in the same order they are
        # visited when performing a depth-first search in pre order
        # on the revisions DAG
        expected_result = [
            "b364f53155044e5308a0f73abb3b5f01995a5b7d",
            "836d498396fb9b5d45c896885f84d8d60a5651dc",
            "ee96c2a2d397b79070d2b6fe3051290963748358",
            "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
            "b94886c500c46e32dc3d7ebae8a5409accd592e5",
            "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
            "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
            "b401c50863475db4440c85c10ac0b6423b61554d",
            "9c5051397e5c2e0c258bb639c3dd34406584ca10",
        ]

        self.check_revisions_ordering("dfs", expected_result, truncated_history=False)

    def test_revisions_walker_dfs_post(self):

        # revisions should be returned in the same order they are
        # visited when performing a depth-first search in post order
        # on the revisions DAG
        expected_result = [
            "b364f53155044e5308a0f73abb3b5f01995a5b7d",
            "b94886c500c46e32dc3d7ebae8a5409accd592e5",
            "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
            "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
            "b401c50863475db4440c85c10ac0b6423b61554d",
            "9c5051397e5c2e0c258bb639c3dd34406584ca10",
            "836d498396fb9b5d45c896885f84d8d60a5651dc",
            "ee96c2a2d397b79070d2b6fe3051290963748358",
            "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
        ]

        self.check_revisions_ordering(
            "dfs_post", expected_result, truncated_history=False
        )

    def test_revisions_walker_bfs(self):

        # revisions should be returned in the same order they are
        # visited when performing a breadth-first search on the
        # revisions DAG
        expected_result = [
            "b364f53155044e5308a0f73abb3b5f01995a5b7d",
            "836d498396fb9b5d45c896885f84d8d60a5651dc",
            "b94886c500c46e32dc3d7ebae8a5409accd592e5",
            "ee96c2a2d397b79070d2b6fe3051290963748358",
            "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
            "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
            "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
            "b401c50863475db4440c85c10ac0b6423b61554d",
            "9c5051397e5c2e0c258bb639c3dd34406584ca10",
        ]

        self.check_revisions_ordering("bfs", expected_result, truncated_history=False)

    def test_revisions_walker_truncated_history(self):

        expected_result = [
            "b364f53155044e5308a0f73abb3b5f01995a5b7d",
            "b94886c500c46e32dc3d7ebae8a5409accd592e5",
            "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
            "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
            "b401c50863475db4440c85c10ac0b6423b61554d",
            "9c5051397e5c2e0c258bb639c3dd34406584ca10",
        ]

        for revs_walker_type in ("committer_date", "bfs", "dfs", "dfs_post"):

            self.check_revisions_ordering(
                revs_walker_type, expected_result, truncated_history=True
            )



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter, deque
from functools import partial
import logging
from typing import Counter as CounterT
from typing import Deque, Dict, Iterable, List, Optional

from swh.model.model import BaseModel
from swh.storage import get_storage

logger = logging.getLogger(__name__)


class RateQueue:
    def __init__(self, size: int, max_errors: int):
        assert size > max_errors
        self._size = size
        self._max_errors = max_errors
        self._errors: Deque[bool] = deque(maxlen=size)

    def add_ok(self, n_ok: int = 1) -> None:
        self._errors.extend([False] * n_ok)

    def add_error(self, n_error: int = 1) -> None:
        self._errors.extend([True] * n_error)

    def limit_reached(self) -> bool:
        return sum(self._errors) > self._max_errors

    def reset(self):
        # mainly for testing purpose
        self._errors.clear()


class TenaciousProxyStorage:
    """Storage proxy that have a tenacious insertion behavior.

    When an xxx_add method is called, it's first attempted as is against the backend
    storage. If a failure occurs, split the list of inserted objects in pieces until
    erroneous objects have been identified, so all the valid objects are guaranteed to
    be inserted.

    Also provides a error-rate limit feature: if more than n errors occurred during the
    insertion of the last p (window_size) objects, stop accepting any insertion.

    The number of insertion retries for a single object can be specified via
    the 'retries' parameter.

    This proxy is mainly intended to be used in a replayer configuration (aka a
    mirror stack), where insertion errors are mostly unexpected (which explains
    the low default ratio errors/window_size).

    Conversely, it should not be used in a loader configuration, as it may
    drop objects without stopping the loader, which leads to holes in the graph.

    Deployments using this proxy should carefully monitor their logs to check any
    failure is expected (because the failed object is corrupted),
    not because of transient errors or issues with the storage backend.

    Sample configuration use case for tenacious storage:

    .. code-block:: yaml

        storage:
          cls: tenacious
        storage:
          cls: remote
          args: http://storage.internal.staging.swh.network:5002/
        error-rate-limit:
          errors: 10
          window_size: 1000

    """

    tenacious_methods: Dict[str, str] = {
        "content_add": "content",
        "content_add_metadata": "content",
        "skipped_content_add": "skipped_content",
        "directory_add": "directory",
        "revision_add": "revision",
        "extid_add": "extid",
        "release_add": "release",
        "snapshot_add": "snapshot",
        "origin_add": "origin",
    }

    def __init__(
        self,
        storage,
        error_rate_limit: Optional[Dict[str, int]] = None,
        retries: int = 3,
    ):
        self.storage = get_storage(**storage)
        if error_rate_limit is None:
            error_rate_limit = {"errors": 10, "window_size": 1000}
        assert "errors" in error_rate_limit
        assert "window_size" in error_rate_limit
        self.rate_queue = RateQueue(
            size=error_rate_limit["window_size"], max_errors=error_rate_limit["errors"],
        )
        self._single_object_retries: int = retries

    def __getattr__(self, key):
        if key in self.tenacious_methods:
            return partial(self._tenacious_add, key)
        return getattr(self.storage, key)

    def _tenacious_add(self, func_name, objects: Iterable[BaseModel]) -> Dict[str, int]:
        """Enqueue objects to write to the storage. This checks if the queue's
           threshold is hit. If it is actually write those to the storage.

        """
        add_function = getattr(self.storage, func_name)
        object_type = self.tenacious_methods[func_name]

        # list of lists of objects; note this to_add list is consumed from the tail
        to_add: List[List[BaseModel]] = [list(objects)]
        n_objs: int = len(to_add[0])

        results: CounterT[str] = Counter()
        retries: int = self._single_object_retries

        while to_add:
            if self.rate_queue.limit_reached():
                logging.error(
                    "Too many insertion errors have been detected; "
                    "disabling insertions"
                )
                raise RuntimeError(
                    "Too many insertion errors have been detected; "
                    "disabling insertions"
                )
            objs = to_add.pop()
            try:
                results.update(add_function(objs))
                self.rate_queue.add_ok(len(objs))
            except Exception as exc:
                if len(objs) > 1:
                    logger.info(
                        f"{func_name}: failed to insert a batch of "
                        f"{len(objs)} {object_type} objects, splitting"
                    )
                    # reinsert objs split in 2 parts at the end of to_add
                    to_add.append(objs[(len(objs) // 2) :])
                    to_add.append(objs[: (len(objs) // 2)])
                    # each time we append a batch in the to_add bag, reset the
                    # one-object-batch retries counter
                    retries = self._single_object_retries
                else:
                    retries -= 1
                    if retries:
                        logger.info(
                            f"{func_name}: failed to insert an {object_type}, retrying"
                        )
                        # give it another chance
                        to_add.append(objs)
                    else:
                        logger.error(
                            f"{func_name}: failed to insert an object, "
                            f"excluding {objs} (from a batch of {n_objs})"
                        )
                        logger.exception(f"Exception was: {exc}")
                        results.update({f"{object_type}:add:errors": 1})
                        self.rate_queue.add_error()
                        # reset the retries counter (needed in case the next
                        # batch is also 1 element only)
                        retries = self._single_object_retries
        return dict(results)

    def reset(self):
        self.rate_queue.reset()



# Copyright (C) 2019-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from os import environ, path

import pytest

from swh.core.db.pytest_plugin import postgresql_fact
import swh.storage
from swh.storage import get_storage
from swh.storage.tests.storage_data import StorageData

SQL_DIR = path.join(path.dirname(swh.storage.__file__), "sql")

environ["LC_ALL"] = "C.UTF-8"


swh_storage_postgresql = postgresql_fact(
    "postgresql_proc", dbname="storage", dump_files=path.join(SQL_DIR, "*.sql")
)


@pytest.fixture
def swh_storage_backend_config(swh_storage_postgresql):
    """Basic pg storage configuration with no journal collaborator
    (to avoid pulling optional dependency on clients of this fixture)

    """
    yield {
        "cls": "postgresql",
        "db": swh_storage_postgresql.dsn,
        "objstorage": {"cls": "memory"},
        "check_config": {"check_write": True},
    }


@pytest.fixture
def swh_storage(swh_storage_backend_config):
    return get_storage(**swh_storage_backend_config)


@pytest.fixture
def sample_data() -> StorageData:
    """Pre-defined sample storage object data to manipulate

    Returns:
        StorageData whose attribute keys are data model objects. Either multiple
        objects: contents, directories, revisions, releases, ... or simple ones:
        content, directory, revision, release, ...

    """
    return StorageData()



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# flake8: noqa
# because of long lines

import copy
import datetime
import json
from unittest.mock import MagicMock, Mock, call

from swh.model.identifiers import CoreSWHID, ExtendedObjectType, ExtendedSWHID
from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    Origin,
    RawExtrinsicMetadata,
)
from swh.storage.migrate_extrinsic_metadata import (
    DEPOSIT_COLS,
    cran_package_from_url,
    handle_row,
)

FETCHER = MetadataFetcher(
    name="migrate-extrinsic-metadata-from-revisions", version="0.0.1",
)
SWH_AUTHORITY = MetadataAuthority(
    type=MetadataAuthorityType.REGISTRY,
    url="https://softwareheritage.org/",
    metadata={},
)
SWH_DEPOSIT_AUTHORITY = MetadataAuthority(
    type=MetadataAuthorityType.DEPOSIT_CLIENT,
    url="https://www.softwareheritage.org",
    metadata={},
)
HAL_AUTHORITY = MetadataAuthority(
    type=MetadataAuthorityType.DEPOSIT_CLIENT,
    url="https://hal.archives-ouvertes.fr/",
    metadata={},
)
INTEL_AUTHORITY = MetadataAuthority(
    type=MetadataAuthorityType.DEPOSIT_CLIENT,
    url="https://software.intel.com",
    metadata={},
)

DIRECTORY_ID = b"a" * 20
DIRECTORY_SWHID = ExtendedSWHID(
    object_type=ExtendedObjectType.DIRECTORY, object_id=DIRECTORY_ID
)


def get_mock_deposit_cur(row_dicts):
    rows = [tuple(d[key] for key in DEPOSIT_COLS) for d in row_dicts]
    deposit_cur = MagicMock()
    deposit_cur.__iter__.side_effect = [iter(rows)]
    return deposit_cur


def test_deposit_1():
    """Has a provider and xmlns, and the metadata is in the revision twice
    (at the root of the metadata dict, and in
    metadata->extrinsic->raw->origin_metadata)"""
    extrinsic_metadata = {
        "title": "Je suis GPL",
        "@xmlns": "http://www.w3.org/2005/Atom",
        "client": "swh",
        "codemeta:url": "https://forge.softwareheritage.org/source/jesuisgpl/",
        "@xmlns:codemeta": "https://doi.org/10.5063/SCHEMA/CODEMETA-2.0",
        "codemeta:author": {
            "codemeta:name": "Stefano Zacchiroli",
            "codemeta:jobTitle": "Maintainer",
        },
        "codemeta:license": {
            "codemeta:url": "https://spdx.org/licenses/GPL-3.0-or-later.html",
            "codemeta:name": "GNU General Public License v3.0 or later",
        },
        # ...
    }
    original_artifacts = [
        {
            "length": 80880,
            "filename": "archive.zip",
            "checksums": {
                "sha1": "bad32a47a359e0e16ebdca2ad2dc6a771dac8f71",
                "sha256": "182b7ee3b7b5b550e83d3bcfed029bb2f625ee760ebfe9557d5fd072bd4e22e4",
            },
        }
    ]

    row = {
        "id": b"\x02#\x10\xdf\x16\xfd\x9eMO\x81\xfe6\xa1B\xe8-\xb9w\xc0\x1d",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(2018, 1, 5, 0, 0, tzinfo=datetime.timezone.utc),
        "committer_date": datetime.datetime(
            2018, 1, 5, 0, 0, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"swh: Deposit 467 in collection swh",
        "metadata": {
            "client": "swh",
            "extrinsic": {
                "raw": {
                    "origin": {
                        "url": "https://www.softwareheritage.org/check-deposit-2020-03-11T11:07:18.424476",
                        "type": "deposit",
                    },
                    "branch_name": "master",
                    "origin_metadata": {
                        "tool": {
                            "name": "swh-deposit",
                            "version": "0.0.1",
                            "configuration": {"sword_version": 2},
                        },
                        "metadata": extrinsic_metadata,
                    },
                },
                "when": "2020-03-11T11:11:36.336283+00:00",
                "provider": "https://deposit.softwareheritage.org/1/private/467/meta/",
            },
            "original_artifact": original_artifacts,
            **extrinsic_metadata,
        },
    }

    origin_url = (
        "https://www.softwareheritage.org/check-deposit-2020-03-11T11:07:18.424476"
    )

    swhid = (
        f"swh:1:dir:ef04a768181417fbc5eef4243e2507915f24deea"
        f";origin={origin_url}"
        f";visit=swh:1:snp:14433c19dbb03ad57c86b58b53a800d6a0e32dd3"
        f";anchor=swh:1:rev:022310df16fd9e4d4f81fe36a142e82db977c01d"
        f";path=/"
    )

    deposit_rows = [
        {
            "deposit.id": 467,
            "deposit.external_id": "check-deposit-2020-03-11T11:07:18.424476",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2020, 3, 11, 11, 7, 18, 688410, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://www.softwareheritage.org",
            "deposit_collection.name": "swh",
            "auth_user.username": "swh",
        },
        {
            "deposit.id": 467,
            "deposit.external_id": "check-deposit-2020-03-11T11:07:18.424476",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2020, 3, 11, 11, 7, 18, 669428, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://www.softwareheritage.org",
            "deposit_collection.name": "swh",
            "auth_user.username": "swh",
        },
    ]

    storage = Mock()

    def origin_get(urls):
        assert urls == [origin_url]
        return [Origin(url=origin_url)]

    storage.origin_get.side_effect = origin_get
    deposit_cur = get_mock_deposit_cur(deposit_rows)
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    deposit_cur.execute.assert_called_once()
    deposit_cur.__iter__.assert_called_once()

    assert storage.method_calls == [
        call.origin_get([origin_url]),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2020, 3, 11, 11, 7, 18, 688410, tzinfo=datetime.timezone.utc
                    ),
                    authority=SWH_DEPOSIT_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:022310df16fd9e4d4f81fe36a142e82db977c01d"
                    ),
                ),
            ]
        ),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2020, 3, 11, 11, 11, 36, 336283, tzinfo=datetime.timezone.utc
                    ),
                    authority=SWH_AUTHORITY,
                    fetcher=FETCHER,
                    format="original-artifacts-json",
                    metadata=json.dumps(original_artifacts).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:022310df16fd9e4d4f81fe36a142e82db977c01d"
                    ),
                ),
            ]
        ),
    ]


def test_deposit_2_without_xmlns():
    """Has a provider, no xmlns, and the metadata is only in
    metadata->extrinsic->raw->origin_metadata)"""
    extrinsic_metadata = {
        "{http://www.w3.org/2005/Atom}id": "hal-01243573",
        "{http://www.w3.org/2005/Atom}author": {
            "{http://www.w3.org/2005/Atom}name": "HAL",
            "{http://www.w3.org/2005/Atom}email": "hal@ccsd.cnrs.fr",
        },
        "{http://www.w3.org/2005/Atom}client": "hal",
        "{http://www.w3.org/2005/Atom}external_identifier": "hal-01243573",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}url": "https://hal-test.archives-ouvertes.fr/hal-01243573",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}name": "The assignment problem",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}author": {
            "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}name": "Morane Gruenpeter"
        },
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}version": 1,
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}identifier": "10.5281/zenodo.438684",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}dateCreated": "2017-11-16T14:54:23+01:00",
    }
    original_artifacts = [
        {
            "length": 208357,
            "filename": "archive.zip",
            "checksums": {
                "sha1": "fa0aec08e8a44ea144dba7ce366c8b5d66c14453",
                "sha256": "f53c05fe947e88ce83751a93bd522b1f88478ea2e7b984c07fc7a7c68128bf87",
            },
        }
    ]

    row = {
        "id": b"\x01\x16\xca\xb7\x19d\xd5\x9c\x85p\xb4\xc5r\x9b(\xbd\xd6<\x9bF",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(
            2018, 1, 17, 12, 54, 0, 723882, tzinfo=datetime.timezone.utc
        ),
        "committer_date": datetime.datetime(
            2018, 1, 17, 12, 54, 0, 723882, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"hal: Deposit 82 in collection hal",
        "metadata": {
            "extrinsic": {
                "raw": {
                    "origin": {
                        "url": "https://hal.archives-ouvertes.fr/hal-01243573",
                        "type": "deposit",
                    },
                    "origin_metadata": {
                        "tool": {
                            "name": "swh-deposit",
                            "version": "0.0.1",
                            "configuration": {"sword_version": 2},
                        },
                        "metadata": extrinsic_metadata,
                        "provider": {
                            "metadata": {},
                            "provider_url": "https://hal.archives-ouvertes.fr/",
                            "provider_name": "hal",
                            "provider_type": "deposit_client",
                        },
                    },
                },
                "when": "2020-05-15T14:27:21.462270+00:00",
                "provider": "https://deposit.softwareheritage.org/1/private/82/meta/",
            },
            "original_artifact": original_artifacts,
        },
    }

    swhid = (
        "swh:1:dir:e04b2a7b8a8838da0693e9fd992a10d6fd211b50"
        ";origin=https://hal.archives-ouvertes.fr/hal-01243573"
        ";visit=swh:1:snp:abc9ae594245a740235b6c039f044352a5f723ec"
        ";anchor=swh:1:rev:0116cab71964d59c8570b4c5729b28bdd63c9b46"
        ";path=/"
    )

    deposit_rows = [
        {
            "deposit.id": 82,
            "deposit.external_id": "hal-01243573",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2018, 1, 17, 12, 54, 1, 533972, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
        {
            "deposit.id": 82,
            "deposit.external_id": "hal-01243573",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2018, 1, 17, 12, 54, 0, 413748, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
    ]

    origin_url = "https://hal.archives-ouvertes.fr/hal-01243573"

    storage = Mock()

    def origin_get(urls):
        assert urls == [origin_url]
        return [Origin(url=origin_url)]

    storage.origin_get.side_effect = origin_get
    deposit_cur = get_mock_deposit_cur(deposit_rows)
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    deposit_cur.execute.assert_called_once()
    deposit_cur.__iter__.assert_called_once()

    assert storage.method_calls == [
        call.origin_get([origin_url]),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2018, 1, 17, 12, 54, 0, 413748, tzinfo=datetime.timezone.utc
                    ),
                    authority=HAL_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json-with-expanded-namespaces",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:0116cab71964d59c8570b4c5729b28bdd63c9b46"
                    ),
                ),
            ]
        ),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2020, 5, 15, 14, 27, 21, 462270, tzinfo=datetime.timezone.utc
                    ),
                    authority=SWH_AUTHORITY,
                    fetcher=FETCHER,
                    format="original-artifacts-json",
                    metadata=json.dumps(original_artifacts).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:0116cab71964d59c8570b4c5729b28bdd63c9b46"
                    ),
                ),
            ]
        ),
    ]


def test_deposit_2_with_xmlns():
    """Has a provider, xmlns, and the metadata is only in
    metadata->extrinsic->raw->origin_metadata)"""
    extrinsic_metadata = {
        "title": "Je suis GPL",
        "@xmlns": "http://www.w3.org/2005/Atom",
        "client": "swh",
        "codemeta:url": "https://forge.softwareheritage.org/source/jesuisgpl/",
        "@xmlns:codemeta": "https://doi.org/10.5063/SCHEMA/CODEMETA-2.0",
        "codemeta:author": {
            "codemeta:name": "Stefano Zacchiroli",
            "codemeta:jobTitle": "Maintainer",
        },
        "codemeta:license": {
            "codemeta:url": "https://spdx.org/licenses/GPL-3.0-or-later.html",
            "codemeta:name": "GNU General Public License v3.0 or later",
        },
        "external_identifier": "je-suis-gpl",
        "codemeta:dateCreated": "2018-01-05",
    }
    original_artifacts = [
        {
            "length": 80880,
            "filename": "archive.zip",
            "checksums": {
                "sha1": "bad32a47a359e0e16ebdca2ad2dc6a771dac8f71",
                "sha256": "182b7ee3b7b5b550e83d3bcfed029bb2f625ee760ebfe9557d5fd072bd4e22e4",
            },
        }
    ]

    row = {
        "id": b'\x01"\x96nP\x93\x17\xae\xcejA\xd0\xf0\x88\xdas<\xc0\x9d\x0f',
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(2018, 1, 5, 0, 0, tzinfo=datetime.timezone.utc),
        "committer_date": datetime.datetime(
            2018, 1, 5, 0, 0, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"swh: Deposit 687 in collection swh",
        "metadata": {
            "extrinsic": {
                "raw": {
                    "origin": {
                        "url": "https://www.softwareheritage.org/check-deposit-2020-06-26T13:50:07.564420",
                        "type": "deposit",
                    },
                    "origin_metadata": {
                        "tool": {
                            "name": "swh-deposit",
                            "version": "0.0.1",
                            "configuration": {"sword_version": 2},
                        },
                        "metadata": extrinsic_metadata,
                        "provider": {
                            "metadata": {},
                            "provider_url": "https://www.softwareheritage.org",
                            "provider_name": "swh",
                            "provider_type": "deposit_client",
                        },
                    },
                },
                "when": "2020-06-26T13:50:22.640625+00:00",
                "provider": "https://deposit.softwareheritage.org/1/private/687/meta/",
            },
            "original_artifact": original_artifacts,
        },
    }

    swhid = (
        "swh:1:dir:ef04a768181417fbc5eef4243e2507915f24deea"
        ";origin=https://www.softwareheritage.org/check-deposit-2020-06-26T13:50:07.564420"
        ";visit=swh:1:snp:8fd469e280fb0724175c64906627f619143d5bdb"
        ";anchor=swh:1:rev:0122966e509317aece6a41d0f088da733cc09d0f"
        ";path=/"
    )
    deposit_rows = [
        {
            "deposit.id": 687,
            "deposit.external_id": "check-deposit-2020-06-26T13:50:07.564420",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2020, 6, 26, 13, 50, 8, 216113, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://www.softwareheritage.org",
            "deposit_collection.name": "swh",
            "auth_user.username": "swh",
        },
        {
            "deposit.id": 687,
            "deposit.external_id": "check-deposit-2020-06-26T13:50:07.564420",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2020, 6, 26, 13, 50, 8, 150498, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://www.softwareheritage.org",
            "deposit_collection.name": "swh",
            "auth_user.username": "swh",
        },
    ]

    origin_url = (
        "https://www.softwareheritage.org/check-deposit-2020-06-26T13:50:07.564420"
    )

    storage = Mock()

    def origin_get(urls):
        assert urls == [origin_url]
        return [Origin(url=origin_url)]

    storage.origin_get.side_effect = origin_get
    deposit_cur = get_mock_deposit_cur(deposit_rows)
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    deposit_cur.execute.assert_called_once()
    deposit_cur.__iter__.assert_called_once()

    assert storage.method_calls == [
        call.origin_get([origin_url]),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2020, 6, 26, 13, 50, 8, 216113, tzinfo=datetime.timezone.utc
                    ),
                    authority=SWH_DEPOSIT_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:0122966e509317aece6a41d0f088da733cc09d0f"
                    ),
                ),
            ]
        ),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2020, 6, 26, 13, 50, 22, 640625, tzinfo=datetime.timezone.utc
                    ),
                    authority=SWH_AUTHORITY,
                    fetcher=FETCHER,
                    format="original-artifacts-json",
                    metadata=json.dumps(original_artifacts).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:0122966e509317aece6a41d0f088da733cc09d0f"
                    ),
                ),
            ]
        ),
    ]


def test_deposit_2_with_json_in_json_and_no_xmlns():
    """New formats introduced in https://forge.softwareheritage.org/D4105 ,
    where the raw metadata is itself JSONed inside the metadata JSON tree
    and https://forge.softwareheritage.org/D4065 where the @xmlns declarations
    are stripped before being sent to the deposit DB"""
    extrinsic_metadata = {
        "id": "hal-02960679",
        "author": {"name": "HAL", "email": "hal@ccsd.cnrs.fr"},
        "client": "hal",
        "codemeta:url": "https://hal.archives-ouvertes.fr/hal-02960679",
        "codemeta:name": "Compressive Spectral Clustering Toolbox",
        "codemeta:author": [
            {"codemeta:name": "Nicolas Tremblay", "codemeta:affiliation": "PANAMA"},
            {"codemeta:name": "Gilles Puy", "codemeta:affiliation": "PANAMA"},
            {"codemeta:name": "R{\\'e}mi Gribonval", "codemeta:affiliation": "PANAMA"},
            {"codemeta:name": "Pierre Vandergheynst"},
        ],
        # ...
    }

    original_artifacts = [
        {
            "url": "https://deposit.softwareheritage.org/1/private/1037/raw/",
            "length": 4546913,
            "filename": "archive.zip",
            "checksums": {
                "sha1": "01a0069c626a383de9a17ace40ecfd588e5c4f26",
                "sha256": "c780a6de91286c70ceecc69fe0c6d201d3fe944aa89e193f3a89ae85dc25c3b1",
            },
        }
    ]

    row = {
        "id": b"J\x9dc{\xa5\x07\xa2\xb93e%\x04(\xe6\xe3\xf0!\xf1\x94\xd0",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(2016, 1, 29, 0, 0, tzinfo=datetime.timezone.utc),
        "committer_date": datetime.datetime(
            2020, 10, 8, 0, 0, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"hal: Deposit 1037 in collection hal",
        "metadata": {
            "extrinsic": {
                "raw": {
                    "origin": {
                        "url": "https://hal.archives-ouvertes.fr/hal-02960679",
                        "type": "deposit",
                    },
                    "origin_metadata": {
                        "tool": {
                            "name": "swh-deposit",
                            "version": "0.2.0",
                            "configuration": {"sword_version": "2"},
                        },
                        "metadata": json.dumps(extrinsic_metadata),
                        "provider": {
                            "metadata": {},
                            "provider_url": "https://hal.archives-ouvertes.fr/",
                            "provider_name": "hal",
                            "provider_type": "deposit_client",
                        },
                    },
                },
                "when": "2020-10-09T13:38:25.888646+00:00",
                "provider": "https://deposit.softwareheritage.org/1/private/1037/meta/",
            },
            "original_artifact": original_artifacts,
        },
    }

    swhid = (
        "swh:1:dir:8bfdf74037ae1c51335995891c6226e0f85e46e2"
        ";origin=https://hal.archives-ouvertes.fr/hal-02960679"
        ";visit=swh:1:snp:bc4a2ddf84dd0cc13d74e1970a1471c2574ed6aa"
        ";anchor=swh:1:rev:4a9d637ba507a2b93365250428e6e3f021f194d0"
        ";path=/"
    )
    deposit_rows = [
        {
            "deposit.id": 1037,
            "deposit.external_id": "hal-02960679",
            "deposit.swhid_context": swhid,
            "deposit.status": "done",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2020, 10, 9, 13, 38, 8, 269611, tzinfo=datetime.timezone.utc,
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
        {
            "deposit.id": 1037,
            "deposit.external_id": "hal-02960679",
            "deposit.swhid_context": swhid,
            "deposit.status": "done",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2020, 10, 9, 13, 38, 7, 394544, tzinfo=datetime.timezone.utc,
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
    ]

    origin_url = "https://hal.archives-ouvertes.fr/hal-02960679"

    storage = Mock()

    def origin_get(urls):
        assert urls == [origin_url]
        return [Origin(url=origin_url)]

    storage.origin_get.side_effect = origin_get
    deposit_cur = get_mock_deposit_cur(deposit_rows)
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    deposit_cur.execute.assert_called_once()
    deposit_cur.__iter__.assert_called_once()

    assert storage.method_calls == [
        call.origin_get([origin_url]),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2020, 10, 9, 13, 38, 7, 394544, tzinfo=datetime.timezone.utc
                    ),
                    authority=HAL_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:4a9d637ba507a2b93365250428e6e3f021f194d0"
                    ),
                ),
            ]
        ),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2020, 10, 9, 13, 38, 25, 888646, tzinfo=datetime.timezone.utc
                    ),
                    authority=SWH_AUTHORITY,
                    fetcher=FETCHER,
                    format="original-artifacts-json",
                    metadata=json.dumps(original_artifacts).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:4a9d637ba507a2b93365250428e6e3f021f194d0"
                    ),
                ),
            ]
        ),
    ]


def test_deposit_3_and_wrong_external_id_in_metadata():
    extrinsic_metadata = {
        "title": "VTune Perf tool",
        "@xmlns": "http://www.w3.org/2005/Atom",
        "client": "swh",
        "codemeta:url": "https://software.intel.com/en-us/vtune",
        "@xmlns:codemeta": "https://doi.org/10.5063/SCHEMA/CODEMETA-2.0",
        "codemeta:author": {
            "codemeta:name": "VTune developer",
            "codemeta:jobTitle": "Software Engineer",
        },
        "external_identifier": "vtune-perf-tool",
        "codemeta:dateCreated": "2019-05-14",
        "codemeta:description": "Modified version of Linux Perf tool which is used by Intel VTune Amplifier",
    }
    source_original_artifacts = [
        {
            "name": "archive.zip",
            "sha1": "07251dbb1d904d143fd7da9935701f17670d4d9b",
            "length": 4350528,
            "sha256": "1f7d111ac79e468002f3edf4b7b2487538d41f6bea362d49b2eb08a537efafb6",
            "sha1_git": "e2d894efcaad4ff36f09eda3b3c0096416b03429",
            "blake2s256": "e2c08b82efbc361fbb2d28aa8352668cd71217f165f63de16b61ed61ace7509d",
            "archive_type": "zip",
        }
    ]
    dest_original_artifacts = [
        {
            "length": 4350528,
            "archive_type": "zip",
            "filename": "archive.zip",
            "checksums": {
                "sha1": "07251dbb1d904d143fd7da9935701f17670d4d9b",
                "sha256": "1f7d111ac79e468002f3edf4b7b2487538d41f6bea362d49b2eb08a537efafb6",
                "sha1_git": "e2d894efcaad4ff36f09eda3b3c0096416b03429",
                "blake2s256": "e2c08b82efbc361fbb2d28aa8352668cd71217f165f63de16b61ed61ace7509d",
            },
        }
    ]

    row = {
        "id": b"\t5`S\xc4\x9a\xd0\xf9\xe6.Q\xc2\x9d>a|y\x11@\xdf",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(2019, 5, 14, 0, 0, tzinfo=datetime.timezone.utc),
        "committer_date": datetime.datetime(
            2019, 5, 14, 0, 0, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"intel: Deposit 268 in collection intel",
        "metadata": {
            **extrinsic_metadata,
            "original_artifact": source_original_artifacts,
        },
    }

    swhid = (
        "swh:1:dir:527c8e4a67d391f2bf1bbc86dd94af5d5cfc8ef7"
        ";origin=https://software.intel.com/f80482de-90a8-4c32-bce4-6f6918d492ff"
        ";visit=swh:1:snp:49d60943d9c061da1aba6266a811412f9db8de2e"
        ";anchor=swh:1:rev:09356053c49ad0f9e62e51c29d3e617c791140df"
        ";path=/"
    )
    deposit_rows = [
        {
            "deposit.id": 268,
            "deposit.external_id": "f80482de-90a8-4c32-bce4-6f6918d492ff",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2019, 5, 14, 7, 49, 36, 775072, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://software.intel.com",
            "deposit_collection.name": "intel",
            "auth_user.username": "intel",
        },
        {
            "deposit.id": 268,
            "deposit.external_id": "f80482de-90a8-4c32-bce4-6f6918d492ff",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2019, 5, 14, 7, 49, 36, 477061, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://software.intel.com",
            "deposit_collection.name": "intel",
            "auth_user.username": "intel",
        },
        {
            "deposit.id": 268,
            "deposit.external_id": "f80482de-90a8-4c32-bce4-6f6918d492ff",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2019, 5, 14, 7, 28, 33, 210100, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://software.intel.com",
            "deposit_collection.name": "intel",
            "auth_user.username": "intel",
        },
        {
            "deposit.id": 268,
            "deposit.external_id": "f80482de-90a8-4c32-bce4-6f6918d492ff",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2019, 5, 14, 7, 28, 33, 41454, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://software.intel.com",
            "deposit_collection.name": "intel",
            "auth_user.username": "intel",
        },
    ]

    origin_url = "https://software.intel.com/f80482de-90a8-4c32-bce4-6f6918d492ff"

    storage = Mock()

    def origin_get(urls):
        assert urls == [origin_url]
        return [Origin(url=origin_url)]

    storage.origin_get.side_effect = origin_get
    deposit_cur = get_mock_deposit_cur(deposit_rows)
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    deposit_cur.execute.assert_called_once()
    deposit_cur.__iter__.assert_called_once()

    assert storage.method_calls == [
        call.origin_get([origin_url]),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2019, 5, 14, 7, 49, 36, 775072, tzinfo=datetime.timezone.utc
                    ),
                    authority=INTEL_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:09356053c49ad0f9e62e51c29d3e617c791140df"
                    ),
                ),
            ]
        ),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2019, 5, 14, 7, 28, 33, 210100, tzinfo=datetime.timezone.utc
                    ),
                    authority=INTEL_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:09356053c49ad0f9e62e51c29d3e617c791140df"
                    ),
                ),
            ]
        ),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2019, 5, 14, 7, 49, 36, 775072, tzinfo=datetime.timezone.utc
                    ),
                    authority=SWH_AUTHORITY,
                    fetcher=FETCHER,
                    format="original-artifacts-json",
                    metadata=json.dumps(dest_original_artifacts).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:09356053c49ad0f9e62e51c29d3e617c791140df"
                    ),
                ),
            ]
        ),
    ]


def test_deposit_3_and_no_swhid():
    extrinsic_metadata = {
        "id": "hal-02337300",
        "@xmlns": "http://www.w3.org/2005/Atom",
        "author": {"name": "HAL", "email": "hal@ccsd.cnrs.fr"},
        "client": "hal",
        "codemeta:url": "https://hal.archives-ouvertes.fr/hal-02337300",
        "codemeta:name": "R package SMM, Simulation and Estimation of Multi-State Discrete-Time Semi-Markov and Markov Models",
        "@xmlns:codemeta": "https://doi.org/10.5063/SCHEMA/CODEMETA-2.0",
        "codemeta:author": [
            # ...
        ],
        # ...
    }
    original_artifacts = [
        # ...
    ]

    row = {
        "id": b"\x91\xe5\xca\x8b'K\xf1\xa8cFd2\xd7Q\xf7A\xbc\x94\xba&",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(2017, 1, 1, 0, 0, tzinfo=datetime.timezone.utc),
        "committer_date": datetime.datetime(
            2019, 11, 6, 14, 47, 30, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"hal: Deposit 342 in collection hal",
        "metadata": {**extrinsic_metadata, "original_artifact": original_artifacts,},
    }
    storage = Mock()

    deposit_cur = None
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    assert storage.method_calls == []


def test_deposit_3_and_unknown_deposit():
    extrinsic_metadata = {
        "title": "Je suis GPL",
        "@xmlns": "http://www.w3.org/2005/Atom",
        "client": "swh",
        "codemeta:url": "https://forge.softwareheritage.org/source/jesuisgpl/",
        "@xmlns:codemeta": "https://doi.org/10.5063/SCHEMA/CODEMETA-2.0",
        "codemeta:author": {
            "codemeta:name": "Stefano Zacchiroli",
            "codemeta:jobTitle": "Maintainer",
        },
        # ...
    }

    row = {
        "id": b"\x8e\x9c\xee\x14\xa6\xad9\xbc\xa44pw\xb8\x7f\xb5\xbb\xd8\x95;\xb1",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(
            2018, 7, 23, 12, 25, 45, 907132, tzinfo=datetime.timezone.utc
        ),
        "committer_date": datetime.datetime(
            2018, 7, 23, 12, 25, 45, 907132, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"swh: Deposit 159 in collection swh",
        "metadata": extrinsic_metadata,
    }

    origin_url = "https://software.intel.com/f80482de-90a8-4c32-bce4-6f6918d492ff"

    storage = Mock()

    deposit_cur = None
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    assert storage.method_calls == []


def test_deposit_4_without_xmlns():
    extrinsic_metadata = {
        "{http://www.w3.org/2005/Atom}id": "hal-01243573",
        "{http://www.w3.org/2005/Atom}author": {
            "{http://www.w3.org/2005/Atom}name": "HAL",
            "{http://www.w3.org/2005/Atom}email": "hal@ccsd.cnrs.fr",
        },
        "{http://www.w3.org/2005/Atom}client": "hal",
        "{http://www.w3.org/2005/Atom}external_identifier": "hal-01243573",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}url": "https://hal-test.archives-ouvertes.fr/hal-01243573",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}name": "The assignment problem",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}author": {
            "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}name": "Morane Gruenpeter"
        },
        # ...
    }

    row = {
        "id": b"\x03\x98\x7f\x05n\xafE\x96\xcd \xd7\xb2\xee\x01\xc9\xb8L\xed\xdf\xa8",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(
            2018, 1, 17, 12, 49, 30, 902891, tzinfo=datetime.timezone.utc
        ),
        "committer_date": datetime.datetime(
            2018, 1, 17, 12, 49, 30, 902891, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b": Deposit 79 in collection hal",
        "metadata": extrinsic_metadata,
    }

    swhid = (
        "swh:1:dir:e04b2a7b8a8838da0693e9fd992a10d6fd211b50"
        ";origin=https://hal.archives-ouvertes.fr/hal-01243573"
        ";visit=swh:1:snp:c31851534c86676a040fb10f438728c90f1c9d55"
        ";anchor=swh:1:rev:43549ebbe70c9cdf0be1647e6319392eaa06f3a3"
        ";path=/"
    )
    deposit_rows = [
        {
            "deposit.id": 79,
            "deposit.external_id": "hal-01243573",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2018, 1, 17, 12, 49, 31, 208347, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
        {
            "deposit.id": 79,
            "deposit.external_id": "hal-01243573",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2018, 1, 17, 12, 49, 30, 645576, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
    ]

    origin_url = "https://hal.archives-ouvertes.fr/hal-01243573"

    storage = Mock()

    def origin_get(urls):
        assert urls == [origin_url]
        return [Origin(url=origin_url)]

    storage.origin_get.side_effect = origin_get
    deposit_cur = get_mock_deposit_cur(deposit_rows)
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    deposit_cur.execute.assert_called_once()
    deposit_cur.__iter__.assert_called_once()

    assert storage.method_calls == [
        call.origin_get([origin_url]),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2018, 1, 17, 12, 49, 30, 645576, tzinfo=datetime.timezone.utc
                    ),
                    authority=HAL_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json-with-expanded-namespaces",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:03987f056eaf4596cd20d7b2ee01c9b84ceddfa8"
                    ),
                ),
            ]
        ),
        # note: no original artifacts
    ]


def test_deposit_4_wrong_origin():
    extrinsic_metadata = {
        "{http://www.w3.org/2005/Atom}id": "hal-01588781",
        "{http://www.w3.org/2005/Atom}author": {
            "{http://www.w3.org/2005/Atom}name": "HAL",
            "{http://www.w3.org/2005/Atom}email": "hal@ccsd.cnrs.fr",
        },
        "{http://www.w3.org/2005/Atom}client": "hal",
        "{http://www.w3.org/2005/Atom}external_identifier": "hal-01588781",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}url": "https://inria.halpreprod.archives-ouvertes.fr/hal-01588781",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}name": "The assignment problem ",
        "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}author": {
            "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}name": "Morane Gruenpeter",
            "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}affiliation": "Initiative pour la Recherche et l'Innovation sur le Logiciel Libre",
        },
        # ...
    }

    row = {
        "id": b"-{\xcec\x1f\xc7\x91\x08\x03\x11\xeb\x83\\GB\x8eXjn\xa4",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(
            2018, 1, 10, 13, 14, 51, 77033, tzinfo=datetime.timezone.utc
        ),
        "committer_date": datetime.datetime(
            2018, 1, 10, 13, 14, 51, 77033, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b": Deposit 75 in collection hal",
        "metadata": extrinsic_metadata,
    }

    swhid = (
        "swh:1:dir:d8971c651fe256942aa4499a3ccdbaa305d3bade"
        ";origin=https://inria.halpreprod.archives-ouvertes.fr/hal-01588781"
        ";visit=swh:1:snp:7c70cc8ea5b79e376605fd6e9b3b04d98861ffc0"
        ";anchor=swh:1:rev:2d7bce631fc791080311eb835c47428e586a6ea4"
        ";path=/"
    )
    deposit_rows = [
        {
            "deposit.id": 75,
            "deposit.external_id": "hal-01588781",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2018, 1, 10, 13, 14, 51, 523963, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
        {
            "deposit.id": 75,
            "deposit.external_id": "hal-01588781",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2018, 1, 10, 13, 14, 50, 555143, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
    ]

    origin_url = "https://inria.halpreprod.archives-ouvertes.fr/hal-01588781"

    storage = Mock()

    def origin_get(urls):
        assert urls == [origin_url]
        return [Origin(url=origin_url)]

    storage.origin_get.side_effect = origin_get
    deposit_cur = get_mock_deposit_cur(deposit_rows)
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    deposit_cur.execute.assert_called_once()
    deposit_cur.__iter__.assert_called_once()

    assert storage.method_calls == [
        call.origin_get([origin_url]),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2018, 1, 10, 13, 14, 50, 555143, tzinfo=datetime.timezone.utc
                    ),
                    authority=HAL_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json-with-expanded-namespaces",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:2d7bce631fc791080311eb835c47428e586a6ea4"
                    ),
                ),
            ]
        ),
        # note: no original artifacts
    ]


def test_deposit_missing_metadata_in_revision():
    extrinsic_metadata = {
        "id": "hal-01243573",
        "@xmlns": "http://www.w3.org/2005/Atom",
        "author": {"name": "HAL", "email": "hal@ccsd.cnrs.fr"},
        "client": "hal",
        "committer": "Administrateur Du Ccsd",
        "codemeta:url": "https://hal-test.archives-ouvertes.fr/hal-01243573",
        "codemeta:name": "The assignment problem",
        "@xmlns:codemeta": "https://doi.org/10.5063/SCHEMA/CODEMETA-2.0",
        "codemeta:author": {"codemeta:name": "Morane Gruenpeter"},
        "codemeta:version": "1",
        "codemeta:identifier": {"#text": "10.5281/zenodo.438684", "@name": "doi",},
        "external_identifier": "hal-01243573",
        "codemeta:dateCreated": "2017-11-16T14:54:23+01:00",
    }
    source_original_artifacts = [
        {
            "name": "archive.zip",
            "sha1": "e8e46324970cd5af7f98c5a86f33f47fa4a41b4a",
            "length": 118650,
            "sha256": "fec81b63d666c43524f966bbd3263da5bee55051d2b48c1659cca5f56fd953e5",
            "sha1_git": "9da2bbd08bec590b36ede2ed43d74cd510b10a79",
            "blake2s256": "5d0973ba3644cc2bcfdb41ff1891744337d6aa9547a7e59fe466f684b027f295",
            "archive_type": "zip",
        }
    ]
    dest_original_artifacts = [
        {
            "length": 118650,
            "archive_type": "zip",
            "filename": "archive.zip",
            "checksums": {
                "sha1": "e8e46324970cd5af7f98c5a86f33f47fa4a41b4a",
                "sha256": "fec81b63d666c43524f966bbd3263da5bee55051d2b48c1659cca5f56fd953e5",
                "sha1_git": "9da2bbd08bec590b36ede2ed43d74cd510b10a79",
                "blake2s256": "5d0973ba3644cc2bcfdb41ff1891744337d6aa9547a7e59fe466f684b027f295",
            },
        }
    ]

    row = {
        "id": b"\x03@v\xf3\xf4\x1e\xe1 N\xb9\xf6@\x82\xcb\xe6\xe9P\xd7\xbb\x8a",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(
            2019, 2, 25, 15, 49, 16, 594536, tzinfo=datetime.timezone.utc
        ),
        "committer_date": datetime.datetime(
            2019, 2, 25, 15, 49, 16, 594536, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"hal: Deposit 229 in collection hal",
        "metadata": {"original_artifact": source_original_artifacts},
    }

    swhid = (
        "swh:1:dir:3d65b6f065118cb856272829b459f0dfa55549aa"
        ";origin=https://hal-test.archives-ouvertes.fr/hal-01243573"
        ";visit=swh:1:snp:322c54ff4023d3216a994bc9ff9ee524ed80ee1f"
        ";anchor=swh:1:rev:034076f3f41ee1204eb9f64082cbe6e950d7bb8a"
        ";path=/"
    )
    deposit_rows = [
        {
            "deposit.id": 229,
            "deposit.external_id": "hal-01243573",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": None,
            "deposit_request.date": datetime.datetime(
                2019, 2, 25, 15, 54, 30, 102072, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
        {
            "deposit.id": 229,
            "deposit.external_id": "hal-01243573",
            "deposit.swhid_context": swhid,
            "deposit.status": "success",
            "deposit_request.metadata": extrinsic_metadata,
            "deposit_request.date": datetime.datetime(
                2019, 2, 25, 15, 49, 12, 302745, tzinfo=datetime.timezone.utc
            ),
            "deposit_client.provider_url": "https://hal.archives-ouvertes.fr/",
            "deposit_collection.name": "hal",
            "auth_user.username": "hal",
        },
    ]

    origin_url = "https://hal.archives-ouvertes.fr/hal-01243573"
    # /!\ not https://hal-test.archives-ouvertes.fr/hal-01243573
    #     do not trust the metadata!

    storage = Mock()

    def origin_get(urls):
        assert urls == [origin_url]
        return [Origin(url=origin_url)]

    storage.origin_get.side_effect = origin_get
    deposit_cur = get_mock_deposit_cur(deposit_rows)
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    deposit_cur.execute.assert_called_once()
    deposit_cur.__iter__.assert_called_once()

    assert storage.method_calls == [
        call.origin_get([origin_url]),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2019, 2, 25, 15, 49, 12, 302745, tzinfo=datetime.timezone.utc
                    ),
                    authority=HAL_AUTHORITY,
                    fetcher=FETCHER,
                    format="sword-v2-atom-codemeta-v2-in-json",
                    metadata=json.dumps(extrinsic_metadata).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:034076f3f41ee1204eb9f64082cbe6e950d7bb8a"
                    ),
                ),
            ]
        ),
        call.raw_extrinsic_metadata_add(
            [
                RawExtrinsicMetadata(
                    target=DIRECTORY_SWHID,
                    discovery_date=datetime.datetime(
                        2019, 2, 25, 15, 54, 30, 102072, tzinfo=datetime.timezone.utc
                    ),
                    authority=SWH_AUTHORITY,
                    fetcher=FETCHER,
                    format="original-artifacts-json",
                    metadata=json.dumps(dest_original_artifacts).encode(),
                    origin=origin_url,
                    revision=CoreSWHID.from_string(
                        "swh:1:rev:034076f3f41ee1204eb9f64082cbe6e950d7bb8a"
                    ),
                ),
            ]
        ),
    ]



# Copyright (C) 2019-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

_use_scylla = bool(os.environ.get("SWH_USE_SCYLLADB", ""))

UDF_LANGUAGE = "lua" if _use_scylla else "java"

if UDF_LANGUAGE == "java":
    # For Cassandra
    CREATE_TABLES_QUERIES = [
        """
        CREATE OR REPLACE FUNCTION ascii_bins_count_sfunc (
            state tuple<int, map<ascii, int>>, -- (nb_none, map<target_type, nb>)
            bin_name ascii
        )
        CALLED ON NULL INPUT
        RETURNS tuple<int, map<ascii, int>>
        LANGUAGE java AS
        $$
            if (bin_name == null) {
                state.setInt(0, state.getInt(0) + 1);
            }
            else {
                Map<String, Integer> counters = state.getMap(
                    1, String.class, Integer.class);
                Integer nb = counters.get(bin_name);
                if (nb == null) {
                    nb = 0;
                }
                counters.put(bin_name, nb + 1);
                state.setMap(1, counters, String.class, Integer.class);
            }
            return state;
        $$;""",
        """
        CREATE OR REPLACE AGGREGATE ascii_bins_count ( ascii )
        SFUNC ascii_bins_count_sfunc
        STYPE tuple<int, map<ascii, int>>
        INITCOND (0, {})
        ;""",
    ]
elif UDF_LANGUAGE == "lua":
    # For ScyllaDB
    # TODO: this is not implementable yet, because ScyllaDB does not support
    # user-defined aggregates. https://github.com/scylladb/scylla/issues/7201
    CREATE_TABLES_QUERIES = []
else:
    assert False, f"{UDF_LANGUAGE} must be 'lua' or 'java'"

CREATE_TABLES_QUERIES = [
    *CREATE_TABLES_QUERIES,
    """
CREATE TYPE IF NOT EXISTS microtimestamp (
    seconds             bigint,
    microseconds        int
);""",
    """
CREATE TYPE IF NOT EXISTS microtimestamp_with_timezone (
    timestamp           frozen<microtimestamp>,
    offset              smallint,
    negative_utc        boolean
);""",
    """
CREATE TYPE IF NOT EXISTS person (
    fullname    blob,
    name        blob,
    email       blob
);""",
    """
CREATE TABLE IF NOT EXISTS content (
    sha1          blob,
    sha1_git      blob,
    sha256        blob,
    blake2s256    blob,
    length        bigint,
    ctime         timestamp,
        -- creation time, i.e. time of (first) injection into the storage
    status        ascii,
    PRIMARY KEY ((sha1, sha1_git, sha256, blake2s256))
);""",
    """
CREATE TABLE IF NOT EXISTS skipped_content (
    sha1          blob,
    sha1_git      blob,
    sha256        blob,
    blake2s256    blob,
    length        bigint,
    ctime         timestamp,
        -- creation time, i.e. time of (first) injection into the storage
    status        ascii,
    reason        text,
    origin        text,
    PRIMARY KEY ((sha1, sha1_git, sha256, blake2s256))
);""",
    """
CREATE TABLE IF NOT EXISTS revision (
    id                              blob PRIMARY KEY,
    date                            microtimestamp_with_timezone,
    committer_date                  microtimestamp_with_timezone,
    type                            ascii,
    directory                       blob,  -- source code "root" directory
    message                         blob,
    author                          person,
    committer                       person,
    synthetic                       boolean,
        -- true iff revision has been created by Software Heritage
    metadata                        text,
        -- extra metadata as JSON(tarball checksums, etc...)
    extra_headers                   frozen<list <list<blob>> >
        -- extra commit information as (tuple(key, value), ...)
);""",
    """
CREATE TABLE IF NOT EXISTS revision_parent (
    id                     blob,
    parent_rank                     int,
        -- parent position in merge commits, 0-based
    parent_id                       blob,
    PRIMARY KEY ((id), parent_rank)
);""",
    """
CREATE TABLE IF NOT EXISTS release
(
    id                              blob PRIMARY KEY,
    target_type                     ascii,
    target                          blob,
    date                            microtimestamp_with_timezone,
    name                            blob,
    message                         blob,
    author                          person,
    synthetic                       boolean,
        -- true iff release has been created by Software Heritage
);""",
    """
CREATE TABLE IF NOT EXISTS directory (
    id              blob PRIMARY KEY,
);""",
    """
CREATE TABLE IF NOT EXISTS directory_entry (
    directory_id    blob,
    name            blob,  -- path name, relative to containing dir
    target          blob,
    perms           int,   -- unix-like permissions
    type            ascii, -- target type
    PRIMARY KEY ((directory_id), name)
);""",
    """
CREATE TABLE IF NOT EXISTS snapshot (
    id              blob PRIMARY KEY,
);""",
    """
-- For a given snapshot_id, branches are sorted by their name,
-- allowing easy pagination.
CREATE TABLE IF NOT EXISTS snapshot_branch (
    snapshot_id     blob,
    name            blob,
    target_type     ascii,
    target          blob,
    PRIMARY KEY ((snapshot_id), name)
);""",
    """
CREATE TABLE IF NOT EXISTS origin_visit (
    origin          text,
    visit           bigint,
    date            timestamp,
    type            text,
    PRIMARY KEY ((origin), visit)
);""",
    """
CREATE TABLE IF NOT EXISTS origin_visit_status (
    origin          text,
    visit           bigint,
    date            timestamp,
    type            text,
    status          ascii,
    metadata        text,
    snapshot        blob,
    PRIMARY KEY ((origin), visit, date)
)
WITH CLUSTERING ORDER BY (visit DESC, date DESC)
;""",  # 'WITH CLUSTERING ORDER BY' is optional with Cassandra 4, but ScyllaDB needs it
    """
CREATE TABLE IF NOT EXISTS origin (
    sha1            blob PRIMARY KEY,
    url             text,
    next_visit_id   int,
        -- We need integer visit ids for compatibility with the pgsql
        -- storage, so we're using lightweight transactions with this trick:
        -- https://stackoverflow.com/a/29391877/539465
);""",
    """
CREATE TABLE IF NOT EXISTS metadata_authority (
    url             text,
    type            ascii,
    PRIMARY KEY ((url), type)
);""",
    """
CREATE TABLE IF NOT EXISTS metadata_fetcher (
    name            ascii,
    version         ascii,
    PRIMARY KEY ((name), version)
);""",
    """
CREATE TABLE IF NOT EXISTS raw_extrinsic_metadata (
    id              blob,

    type            text,
    target          text,

    -- metadata source
    authority_type  text,
    authority_url   text,
    discovery_date  timestamp,
    fetcher_name    ascii,
    fetcher_version ascii,

    -- metadata itself
    format          ascii,
    metadata        blob,

    -- context
    origin          text,
    visit           bigint,
    snapshot        text,
    release         text,
    revision        text,
    path            blob,
    directory       text,

    PRIMARY KEY ((target), authority_type, authority_url, discovery_date, id)

    -- An explanation is in order for this primary key:
    --
    -- Intuitively, the primary key should only be 'id', because two metadata
    -- entries are the same iff the id is the same; and 'id' is used for
    -- deduplication.
    --
    -- However, we also want to query by
    -- (target, authority_type, authority_url, discovery_date)
    -- The naive solution to this would be an extra table, to use as index;
    -- but it means 1. extra code to keep them in sync 2. overhead when writing
    -- 3. overhead + random reads (instead of linear) when reading.
    --
    -- Therefore, we use a single table for both, by adding the column
    -- we want to query with before the id.
    -- It solves both a) the query/order issues and b) the uniqueness issue because:
    --
    -- a) adding the id at the end of the primary key does not change the rows' order:
    --    for two different rows, id1 != id2, so
    --    (target1, ..., date1) < (target2, ..., date2)
    --    <=> (target1, ..., date1, id1) < (target2, ..., date2, id2)
    --
    -- b) the id is a hash of all the columns, so:
    --    rows are the same
    --    <=> id1 == id2
    --    <=> (target1, ..., date1, id1) == (target2, ..., date2, id2)
);""",
    """
CREATE TABLE IF NOT EXISTS raw_extrinsic_metadata_by_id (
    id              blob,
    target          text,
    authority_type  text,
    authority_url   text,
    PRIMARY KEY ((id))
);""",
    """
CREATE TABLE IF NOT EXISTS object_count (
    partition_key   smallint,  -- Constant, must always be 0
    object_type     ascii,
    count           counter,
    PRIMARY KEY ((partition_key), object_type)
);""",
    """
CREATE TABLE IF NOT EXISTS extid (
    extid_type      ascii,
    extid           blob,
    target_type     ascii,
    target          blob,
    PRIMARY KEY ((extid_type, extid), target_type, target)
);""",
    """
CREATE TABLE IF NOT EXISTS extid_by_target (
    target_type     ascii,
    target          blob,
    target_token    bigint, -- value of token(pk) on the "primary" table
    PRIMARY KEY ((target_type, target), target_token)
);""",
]

CONTENT_INDEX_TEMPLATE = """
-- Secondary table, used for looking up "content" from a single hash
CREATE TABLE IF NOT EXISTS content_by_{main_algo} (
    {main_algo}   blob,
    target_token  bigint, -- value of token(pk) on the "primary" table
    PRIMARY KEY (({main_algo}), target_token)
);

CREATE TABLE IF NOT EXISTS skipped_content_by_{main_algo} (
    {main_algo}   blob,
    target_token  bigint, -- value of token(pk) on the "primary" table
    PRIMARY KEY (({main_algo}), target_token)
);
"""

TABLES = [
    "skipped_content",
    "content",
    "revision",
    "revision_parent",
    "release",
    "directory",
    "directory_entry",
    "snapshot",
    "snapshot_branch",
    "origin_visit",
    "origin",
    "raw_extrinsic_metadata",
    "object_count",
    "origin_visit_status",
    "metadata_authority",
    "metadata_fetcher",
    "extid",
    "extid_by_target",
]

HASH_ALGORITHMS = ["sha1", "sha1_git", "sha256", "blake2s256"]

for main_algo in HASH_ALGORITHMS:
    CREATE_TABLES_QUERIES.extend(
        CONTENT_INDEX_TEMPLATE.format(
            main_algo=main_algo,
            other_algos=", ".join(
                [algo for algo in HASH_ALGORITHMS if algo != main_algo]
            ),
        ).split("\n\n")
    )

    TABLES.append("content_by_%s" % main_algo)
    TABLES.append("skipped_content_by_%s" % main_algo)



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os
from typing import Any, Dict, Optional

from swh.core import config
from swh.core.api import RPCServerApp
from swh.core.api import encode_data_server as encode_data
from swh.core.api import error_handler
from swh.storage import get_storage as get_swhstorage

from ..exc import StorageArgumentException
from ..interface import StorageInterface
from ..metrics import timed
from .serializers import DECODERS, ENCODERS


def get_storage():
    global storage
    if not storage:
        storage = get_swhstorage(**app.config["storage"])

    return storage


class StorageServerApp(RPCServerApp):
    extra_type_decoders = DECODERS
    extra_type_encoders = ENCODERS


app = StorageServerApp(
    __name__, backend_class=StorageInterface, backend_factory=get_storage
)
storage = None


@app.errorhandler(StorageArgumentException)
def argument_error_handler(exception):
    return error_handler(exception, encode_data, status_code=400)


@app.errorhandler(Exception)
def my_error_handler(exception):
    return error_handler(exception, encode_data)


@app.route("/")
@timed
def index():
    return """<html>
<head><title>Software Heritage storage server</title></head>
<body>
<p>You have reached the
<a href="https://www.softwareheritage.org/">Software Heritage</a>
storage server.<br />
See its
<a href="https://docs.softwareheritage.org/devel/swh-storage/">documentation
and API</a> for more information</p>
</body>
</html>"""


@app.route("/stat/counters", methods=["GET"])
@timed
def stat_counters():
    return encode_data(get_storage().stat_counters())


@app.route("/stat/refresh", methods=["GET"])
@timed
def refresh_stat_counters():
    return encode_data(get_storage().refresh_stat_counters())


api_cfg = None


def load_and_check_config(config_path: Optional[str]) -> Dict[str, Any]:
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_path: Path to the configuration file to load

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_path:
        raise EnvironmentError("Configuration file must be defined")

    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Configuration file {config_path} does not exist")

    cfg = config.read(config_path)
    if "storage" not in cfg:
        raise KeyError("Missing 'storage' configuration")

    return cfg


def make_app_from_configfile() -> StorageServerApp:
    """Run the WSGI app from the webserver, loading the configuration from
       a configuration file.

       SWH_CONFIG_FILENAME environment variable defines the
       configuration path to load.

    """
    global api_cfg
    if not api_cfg:
        config_path = os.environ.get("SWH_CONFIG_FILENAME")
        api_cfg = load_and_check_config(config_path)
        app.config.update(api_cfg)
    handler = logging.StreamHandler()
    app.logger.addHandler(handler)
    return app


if __name__ == "__main__":
    print("Deprecated. Use swh-storage")



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from typing import Dict, Iterable, List, Set

from swh.model.model import Content, Directory, Revision, Sha1Git, SkippedContent
from swh.storage import get_storage
from swh.storage.interface import StorageInterface


class FilteringProxyStorage:
    """Filtering Storage implementation. This is in charge of transparently
       filtering out known objects prior to adding them to storage.

    Sample configuration use case for filtering storage:

    .. code-block: yaml

        storage:
          cls: filter
          storage:
            cls: remote
            url: http://storage.internal.staging.swh.network:5002/

    """

    object_types = ["content", "skipped_content", "directory", "revision"]

    def __init__(self, storage):
        self.storage: StorageInterface = get_storage(**storage)

    def __getattr__(self, key):
        if key == "storage":
            raise AttributeError(key)
        return getattr(self.storage, key)

    def content_add(self, content: List[Content]) -> Dict[str, int]:
        contents_to_add = self._filter_missing_contents(content)
        return self.storage.content_add(
            [x for x in content if x.sha256 in contents_to_add]
        )

    def skipped_content_add(self, content: List[SkippedContent]) -> Dict[str, int]:
        contents_to_add = self._filter_missing_skipped_contents(content)
        return self.storage.skipped_content_add(
            [x for x in content if x.sha1_git is None or x.sha1_git in contents_to_add]
        )

    def directory_add(self, directories: List[Directory]) -> Dict[str, int]:
        missing_ids = self._filter_missing_ids("directory", (d.id for d in directories))
        return self.storage.directory_add(
            [d for d in directories if d.id in missing_ids]
        )

    def revision_add(self, revisions: List[Revision]) -> Dict[str, int]:
        missing_ids = self._filter_missing_ids("revision", (r.id for r in revisions))
        return self.storage.revision_add([r for r in revisions if r.id in missing_ids])

    def _filter_missing_contents(self, contents: List[Content]) -> Set[bytes]:
        """Return only the content keys missing from swh

        Args:
            content_hashes: List of sha256 to check for existence in swh
                storage

        """
        missing_contents = []
        for content in contents:
            missing_contents.append(content.hashes())

        return set(self.storage.content_missing(missing_contents, key_hash="sha256",))

    def _filter_missing_skipped_contents(
        self, contents: List[SkippedContent]
    ) -> Set[Sha1Git]:
        """Return only the content keys missing from swh

        Args:
            content_hashes: List of sha1_git to check for existence in swh
                storage

        """
        missing_contents = [c.hashes() for c in contents if c.sha1_git is not None]

        ids = set()
        for c in self.storage.skipped_content_missing(missing_contents):
            if c is None or c.get("sha1_git") is None:
                continue
            ids.add(c["sha1_git"])
        return ids

    def _filter_missing_ids(self, object_type: str, ids: Iterable[bytes]) -> Set[bytes]:
        """Filter missing ids from the storage for a given object type.

        Args:
            object_type: object type to use {revision, directory}
            ids: List of object_type ids

        Returns:
            Missing ids from the storage for object_type

        """
        missing_ids = []
        for id in ids:
            missing_ids.append(id)

        fn_by_object_type = {
            "revision": self.storage.revision_missing,
            "directory": self.storage.directory_missing,
        }

        fn = fn_by_object_type[object_type]
        return set(fn(missing_ids))



#!/usr/bin/env python3

# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""This is an executable script to migrate extrinsic revision metadata from
the revision table to the new extrinsic metadata storage.

This is designed to be as conservative as possible, following this principle:
for each revision the script reads (in "handle_row"), it will read some of the
fields, write them directly to the metadata storage, and remove them.
Then it checks all the remaining fields are in a hardcoded list of fields that
are known not to require migration.

This means that every field that isn't migrated was explicitly reviewed while
writing this script.

Additionally, this script contains many assertions to prevent false positives
in its heuristics.
"""

import datetime
import hashlib
import json
import os
import re
import sys
import time
from typing import Any, Dict, Optional
from urllib.error import HTTPError
from urllib.parse import unquote, urlparse
from urllib.request import urlopen

import iso8601
import psycopg2

from swh.core.db import BaseDb
from swh.model.hashutil import hash_to_hex
from swh.model.identifiers import (
    CoreSWHID,
    ExtendedObjectType,
    ExtendedSWHID,
    ObjectType,
    QualifiedSWHID,
)
from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    RawExtrinsicMetadata,
    Sha1Git,
)
from swh.storage import get_storage
from swh.storage.algos.origin import iter_origin_visit_statuses, iter_origin_visits
from swh.storage.algos.snapshot import snapshot_get_all_branches

# XML namespaces and fields for metadata coming from the deposit:

CODEMETA_NS = "https://doi.org/10.5063/SCHEMA/CODEMETA-2.0"
ATOM_NS = "http://www.w3.org/2005/Atom"
ATOM_KEYS = ["id", "author", "external_identifier", "title"]

# columns of the revision table (of the storage DB)
REVISION_COLS = [
    "id",
    "directory",
    "date",
    "committer_date",
    "type",
    "message",
    "metadata",
]

# columns of the tables of the deposit DB
DEPOSIT_COLS = [
    "deposit.id",
    "deposit.external_id",
    "deposit.swhid_context",
    "deposit.status",
    "deposit_request.metadata",
    "deposit_request.date",
    "deposit_client.provider_url",
    "deposit_collection.name",
    "auth_user.username",
]

# Formats we write to the extrinsic metadata storage
OLD_DEPOSIT_FORMAT = (
    "sword-v2-atom-codemeta-v2-in-json-with-expanded-namespaces"  # before february 2018
)
NEW_DEPOSIT_FORMAT = "sword-v2-atom-codemeta-v2-in-json"  # after february 2018
GNU_FORMAT = "gnu-tree-json"
NIXGUIX_FORMAT = "nixguix-sources-json"
NPM_FORMAT = "replicate-npm-package-json"
ORIGINAL_ARTIFACT_FORMAT = "original-artifacts-json"
PYPI_FORMAT = "pypi-project-json"

# Information about this script, for traceability
FETCHER = MetadataFetcher(
    name="migrate-extrinsic-metadata-from-revisions", version="0.0.1",
)

# Authorities that we got the metadata from
AUTHORITIES = {
    "npmjs": MetadataAuthority(
        type=MetadataAuthorityType.FORGE, url="https://npmjs.com/", metadata={}
    ),
    "pypi": MetadataAuthority(
        type=MetadataAuthorityType.FORGE, url="https://pypi.org/", metadata={}
    ),
    "gnu": MetadataAuthority(
        type=MetadataAuthorityType.FORGE, url="https://ftp.gnu.org/", metadata={}
    ),
    "swh": MetadataAuthority(
        type=MetadataAuthorityType.REGISTRY,
        url="https://softwareheritage.org/",
        metadata={},
    ),  # for original_artifact (which are checksums computed by SWH)
}

# Regular expression for the format of revision messages written by the
# deposit loader
deposit_revision_message_re = re.compile(
    b"(?P<client>[a-z-]*): "
    b"Deposit (?P<deposit_id>[0-9]+) in collection (?P<collection>[a-z-]+).*"
)


# not reliable, because PyPI allows arbitrary names
def pypi_project_from_filename(filename):
    original_filename = filename
    if filename.endswith(".egg"):
        return None
    elif filename == "mongomotor-0.13.0.n.tar.gz":
        return "mongomotor"
    elif re.match(r"datahaven-rev[0-9]+\.tar\.gz", filename):
        return "datahaven"
    elif re.match(r"Dtls-[0-9]\.[0-9]\.[0-9]\.sdist_with_openssl\..*", filename):
        return "Dtls"
    elif re.match(r"(gae)?pytz-20[0-9][0-9][a-z]\.(tar\.gz|zip)", filename):
        return filename.split("-", 1)[0]
    elif filename.startswith(("powny-", "obedient.powny-",)):
        return filename.split("-")[0]
    elif filename.startswith("devpi-theme-16-"):
        return "devpi-theme-16"
    elif re.match("[^-]+-[0-9]+.tar.gz", filename):
        return filename.split("-")[0]
    elif filename == "ohai-1!0.tar.gz":
        return "ohai"
    elif filename == "collective.topicitemsevent-0.1dvl.tar.gz":
        return "collective.topicitemsevent"
    elif filename.startswith(
        ("SpiNNStorageHandlers-1!", "sPyNNakerExternalDevicesPlugin-1!")
    ):
        return filename.split("-")[0]
    elif filename.startswith("limnoria-201"):
        return "limnoria"
    elif filename.startswith("pytz-20"):
        return "pytz"
    elif filename.startswith("youtube_dl_server-alpha."):
        return "youtube_dl_server"
    elif filename == "json-extensions-b76bc7d.tar.gz":
        return "json-extensions"
    elif filename == "LitReview-0.6989ev.tar.gz":
        # typo of "dev"
        return "LitReview"
    elif filename.startswith("django_options-r"):
        return "django_options"
    elif filename == "Greater than, equal, or less Library-0.1.tar.gz":
        return "Greater-than-equal-or-less-Library"
    elif filename.startswith("upstart--main-"):
        return "upstart"
    elif filename == "duckduckpy0.1.tar.gz":
        return "duckduckpy"
    elif filename == "QUI for MPlayer snapshot_9-14-2011.zip":
        return "QUI-for-MPlayer"
    elif filename == "Eddy's Memory Game-1.0.zip":
        return "Eddy-s-Memory-Game"
    elif filename == "jekyll2nikola-0-0-1.tar.gz":
        return "jekyll2nikola"
    elif filename.startswith("ore.workflowed"):
        return "ore.workflowed"
    elif re.match("instancemanager-[0-9]*", filename):
        return "instancemanager"
    elif filename == "OrzMC_W&L-1.0.0.tar.gz":
        return "OrzMC-W-L"
    elif filename == "use0mk.tar.gz":
        return "use0mk"
    elif filename == "play-0-develop-1-gd67cd85.tar.gz":
        return "play"
    filename = filename.replace(" ", "-")

    match = re.match(
        r"^(?P<project_name>[a-z_.-]+)"  # project name
        r"\.(tar\.gz|tar\.bz2|tgz|zip)$",  # extension
        filename,
        re.I,
    )
    if match:
        return match.group("project_name")

    # First try with a rather strict format, but that allows accidentally
    # matching the version as part of the package name
    match = re.match(
        r"^(?P<project_name>[a-z0-9_.]+?([-_][a-z][a-z0-9.]+?)*?)"  # project name
        r"-v?"
        r"([0-9]+!)?"  # epoch
        r"[0-9_.]+([a-z]+[0-9]+)?"  # "main" version
        r"([.-]?(alpha|beta|dev|post|pre|rc)(\.?[0-9]+)?)*"  # development status
        r"([.-]?20[012][0-9]{5,9})?"  # date
        r"([.-]g?[0-9a-f]+)?"  # git commit
        r"([-+]py(thon)?(3k|[23](\.?[0-9]{1,2})?))?"  # python version
        r"\.(tar\.gz|tar\.bz2|tgz|zip)$",  # extension
        filename,
        re.I,
    )
    if match:
        return match.group("project_name")

    # If that doesn't work, give up on trying to parse version suffixes,
    # and just find the first version-like occurrence in the file name

    match = re.match(
        r"^(?P<project_name>[a-z0-9_.-]+?)"  # project name
        r"[-_.]v?"
        r"([0-9]+!)?"  # epoch
        r"("  # "main" version
        r"[0-9_]+\.[0-9_.]+([a-z]+[0-9]+)?"  # classic version number
        r"|20[012][0-9]{5,9}"  # date as integer
        r"|20[012][0-9]-[01][0-9]-[0-3][0-9]"  # date as ISO 8601
        r")"  # end of "main" version
        r"[a-z]?(dev|pre)?"  # direct version suffix
        r"([._-].*)?"  # extra suffixes
        r"\.(tar\.gz|tar\.bz2|tgz|zip)$",  # extension
        filename,
        re.I,
    )
    if match:
        return match.group("project_name")

    # If that still doesn't work, give one last chance if there's only one
    # dash or underscore in the name

    match = re.match(
        r"^(?P<project_name>[^_-]+)"  # project name
        r"[_-][^_-]+"  # version
        r"\.(tar\.gz|tar\.bz2|tgz|zip)$",  # extension
        filename,
    )
    assert match, original_filename
    return match.group("project_name")


def pypi_origin_from_project_name(project_name: str) -> str:
    return f"https://pypi.org/project/{project_name}/"


def pypi_origin_from_filename(storage, rev_id: bytes, filename: str) -> Optional[str]:
    project_name = pypi_project_from_filename(filename)
    origin = pypi_origin_from_project_name(project_name)
    # But unfortunately, the filename is user-provided, and doesn't
    # necessarily match the package name on pypi. Therefore, we need
    # to check it.
    if _check_revision_in_origin(storage, origin, rev_id):
        return origin

    # if the origin we guessed does not exist, query the PyPI API with the
    # project name we guessed. If only the capitalisation and dash/underscores
    # are wrong (by far the most common case), PyPI kindly corrects them.
    try:
        resp = urlopen(f"https://pypi.org/pypi/{project_name}/json/")
    except HTTPError as e:
        assert e.code == 404
        # nope; PyPI couldn't correct the wrong project name
        return None
    assert resp.code == 200, resp.code
    project_name = json.load(resp)["info"]["name"]
    origin = pypi_origin_from_project_name(project_name)

    if _check_revision_in_origin(storage, origin, rev_id):
        return origin
    else:
        # The origin exists, but the revision does not belong in it.
        # This happens sometimes, as the filename we guessed the origin
        # from is user-provided.
        return None


def cran_package_from_url(filename):
    match = re.match(
        r"^https://cran\.r-project\.org/src/contrib/"
        r"(?P<package_name>[a-zA-Z0-9.]+)_[0-9.-]+(\.tar\.gz)?$",
        filename,
    )
    assert match, filename
    return match.group("package_name")


def npm_package_from_source_url(package_source_url):
    match = re.match(
        "^https://registry.npmjs.org/(?P<package_name>.*)/-/[^/]+.tgz$",
        package_source_url,
    )
    assert match, package_source_url
    return unquote(match.group("package_name"))


def remove_atom_codemeta_metadata_with_xmlns(metadata):
    """Removes all known Atom and Codemeta metadata fields from the dict,
    assuming this is a dict generated by xmltodict without expanding namespaces.
    """
    keys_to_remove = ATOM_KEYS + ["@xmlns", "@xmlns:codemeta"]
    for key in list(metadata):
        if key.startswith("codemeta:") or key in keys_to_remove:
            del metadata[key]


def remove_atom_codemeta_metadata_without_xmlns(metadata):
    """Removes all known Atom and Codemeta metadata fields from the dict,
    assuming this is a dict generated by xmltodict with expanded namespaces.
    """
    for key in list(metadata):
        if key.startswith(("{%s}" % ATOM_NS, "{%s}" % CODEMETA_NS)):
            del metadata[key]


def _check_revision_in_origin(storage, origin, revision_id):
    seen_snapshots = set()  # no need to visit them again
    seen_revisions = set()

    for visit in iter_origin_visits(storage, origin):
        for status in iter_origin_visit_statuses(storage, origin, visit.visit):
            if status.snapshot is None:
                continue
            if status.snapshot in seen_snapshots:
                continue
            seen_snapshots.add(status.snapshot)
            snapshot = snapshot_get_all_branches(storage, status.snapshot)
            for (branch_name, branch) in snapshot.branches.items():
                if branch is None:
                    continue

                # If it's the revision passed as argument, then it is indeed in the
                # origin
                if branch.target == revision_id:
                    return True

                # Else, let's make sure the branch doesn't have any other revision

                # Get the revision at the top of the branch.
                if branch.target in seen_revisions:
                    continue
                seen_revisions.add(branch.target)
                revision = storage.revision_get([branch.target])[0]

                if revision is None:
                    # https://forge.softwareheritage.org/T997
                    continue

                # Check it doesn't have parents (else we would have to
                # recurse)
                assert revision.parents == (), "revision with parents"

    return False


def debian_origins_from_row(row, storage):
    """Guesses a Debian origin from a row. May return an empty list if it
    cannot reliably guess it, but all results are guaranteed to be correct."""
    filenames = [entry["filename"] for entry in row["metadata"]["original_artifact"]]
    package_names = {filename.split("_")[0] for filename in filenames}
    assert len(package_names) == 1, package_names
    (package_name,) = package_names

    candidate_origins = [
        f"deb://Debian/packages/{package_name}",
        f"deb://Debian-Security/packages/{package_name}",
        f"http://snapshot.debian.org/package/{package_name}/",
    ]

    return [
        origin
        for origin in candidate_origins
        if _check_revision_in_origin(storage, origin, row["id"])
    ]


# Cache of origins that are known to exist
_origins = set()


def assert_origin_exists(storage, origin):
    assert check_origin_exists(storage, origin), origin


def check_origin_exists(storage, origin):
    return (
        (
            hashlib.sha1(origin.encode()).digest() in _origins  # very fast
            or storage.origin_get([origin])[0] is not None  # slow, but up to date
        ),
        origin,
    )


def load_metadata(
    storage,
    revision_id,
    directory_id,
    discovery_date: datetime.datetime,
    metadata: Dict[str, Any],
    format: str,
    authority: MetadataAuthority,
    origin: Optional[str],
    dry_run: bool,
):
    """Does the actual loading to swh-storage."""
    directory_swhid = ExtendedSWHID(
        object_type=ExtendedObjectType.DIRECTORY, object_id=directory_id
    )
    revision_swhid = CoreSWHID(object_type=ObjectType.REVISION, object_id=revision_id)
    obj = RawExtrinsicMetadata(
        target=directory_swhid,
        discovery_date=discovery_date,
        authority=authority,
        fetcher=FETCHER,
        format=format,
        metadata=json.dumps(metadata).encode(),
        origin=origin,
        revision=revision_swhid,
    )
    if not dry_run:
        storage.raw_extrinsic_metadata_add([obj])


def handle_deposit_row(
    row,
    discovery_date: Optional[datetime.datetime],
    origin,
    storage,
    deposit_cur,
    dry_run: bool,
):
    """Loads metadata from the deposit database (which is more reliable as the
    metadata on the revision object, as some versions of the deposit loader were
    a bit lossy; and they used very different format for the field in the
    revision table).
    """
    parsed_message = deposit_revision_message_re.match(row["message"])
    assert parsed_message is not None, row["message"]

    deposit_id = int(parsed_message.group("deposit_id"))
    collection = parsed_message.group("collection").decode()
    client_name = parsed_message.group("client").decode()

    deposit_cur.execute(
        f"SELECT {', '.join(DEPOSIT_COLS)} FROM deposit "
        f"INNER JOIN deposit_collection "
        f" ON (deposit.collection_id=deposit_collection.id) "
        f"INNER JOIN deposit_client ON (deposit.client_id=deposit_client.user_ptr_id) "
        f"INNER JOIN auth_user ON (deposit.client_id=auth_user.id) "
        f"INNER JOIN deposit_request ON (deposit.id=deposit_request.deposit_id) "
        f"WHERE deposit.id = %s",
        (deposit_id,),
    )

    provider_urls = set()
    swhids = set()
    metadata_entries = []
    dates = set()
    external_identifiers = set()
    for deposit_request_row in deposit_cur:
        deposit_request = dict(zip(DEPOSIT_COLS, deposit_request_row))

        # Sanity checks to make sure we selected the right deposit
        assert deposit_request["deposit.id"] == deposit_id
        assert deposit_request["deposit_collection.name"] == collection, deposit_request
        if client_name != "":
            # Sometimes it's missing from the commit message
            assert deposit_request["auth_user.username"] == client_name

        # Date of the deposit request (either the initial request, of subsequent ones)
        date = deposit_request["deposit_request.date"]
        dates.add(date)

        assert deposit_request["deposit.swhid_context"], deposit_request
        external_identifiers.add(deposit_request["deposit.external_id"])
        swhids.add(deposit_request["deposit.swhid_context"])

        # Client of the deposit
        provider_urls.add(deposit_request["deposit_client.provider_url"])

        metadata = deposit_request["deposit_request.metadata"]
        if metadata is not None:
            json.dumps(metadata).encode()  # check it's valid
            if "@xmlns" in metadata:
                assert metadata["@xmlns"] == ATOM_NS
                assert metadata["@xmlns:codemeta"] in (CODEMETA_NS, [CODEMETA_NS])
                format = NEW_DEPOSIT_FORMAT
            elif "{http://www.w3.org/2005/Atom}id" in metadata:
                assert (
                    "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}author" in metadata
                    or "{http://www.w3.org/2005/Atom}author" in metadata
                )
                format = OLD_DEPOSIT_FORMAT
            else:
                # new format introduced in
                # https://forge.softwareheritage.org/D4065
                # it's the same as the first case, but with the @xmlns
                # declarations stripped
                # Most of them should have the "id", but some revisions,
                # like 4d3890004fade1f4ec3bf7004a4af0c490605128, are missing
                # this field
                assert "id" in metadata or "title" in metadata
                assert "codemeta:author" in metadata
                format = NEW_DEPOSIT_FORMAT
            metadata_entries.append((date, format, metadata))

    if discovery_date is None:
        discovery_date = max(dates)

    # Sanity checks to make sure deposit requests are consistent with each other
    assert len(metadata_entries) >= 1, deposit_id
    assert len(provider_urls) == 1, f"expected 1 provider url, got {provider_urls}"
    (provider_url,) = provider_urls
    assert len(swhids) == 1
    (swhid,) = swhids
    assert (
        len(external_identifiers) == 1
    ), f"expected 1 external identifier, got {external_identifiers}"
    (external_identifier,) = external_identifiers

    # computed the origin from the external_identifier if we don't have one
    if origin is None:
        origin = f"{provider_url.strip('/')}/{external_identifier}"

        # explicit list of mistakes that happened in the past, but shouldn't
        # happen again:
        if origin == "https://hal.archives-ouvertes.fr/hal-01588781":
            # deposit id 75
            origin = "https://inria.halpreprod.archives-ouvertes.fr/hal-01588781"
        elif origin == "https://hal.archives-ouvertes.fr/hal-01588782":
            # deposit id 76
            origin = "https://inria.halpreprod.archives-ouvertes.fr/hal-01588782"
        elif origin == "https://hal.archives-ouvertes.fr/hal-01592430":
            # deposit id 143
            origin = "https://hal-preprod.archives-ouvertes.fr/hal-01592430"
        elif origin == "https://hal.archives-ouvertes.fr/hal-01588927":
            origin = "https://inria.halpreprod.archives-ouvertes.fr/hal-01588927"
        elif origin == "https://hal.archives-ouvertes.fr/hal-01593875":
            # deposit id 175
            origin = "https://hal-preprod.archives-ouvertes.fr/hal-01593875"
        elif deposit_id == 160:
            assert origin == "https://www.softwareheritage.org/je-suis-gpl", origin
            origin = "https://forge.softwareheritage.org/source/jesuisgpl/"
        elif origin == "https://hal.archives-ouvertes.fr/hal-01588942":
            # deposit id 90
            origin = "https://inria.halpreprod.archives-ouvertes.fr/hal-01588942"
        elif origin == "https://hal.archives-ouvertes.fr/hal-01592499":
            # deposit id 162
            origin = "https://hal-preprod.archives-ouvertes.fr/hal-01592499"
        elif origin == "https://hal.archives-ouvertes.fr/hal-01588935":
            # deposit id 89
            origin = "https://hal-preprod.archives-ouvertes.fr/hal-01588935"

        assert_origin_exists(storage, origin)

    # check the origin we computed matches the one in the deposit db
    swhid_origin = QualifiedSWHID.from_string(swhid).origin
    if origin is not None:
        # explicit list of mistakes that happened in the past, but shouldn't
        # happen again:
        exceptions = [
            (
                # deposit id 229
                "https://hal.archives-ouvertes.fr/hal-01243573",
                "https://hal-test.archives-ouvertes.fr/hal-01243573",
            ),
            (
                # deposit id 199
                "https://hal.archives-ouvertes.fr/hal-01243065",
                "https://hal-test.archives-ouvertes.fr/hal-01243065",
            ),
            (
                # deposit id 164
                "https://hal.archives-ouvertes.fr/hal-01593855",
                "https://hal-preprod.archives-ouvertes.fr/hal-01593855",
            ),
        ]
        if (origin, swhid_origin) not in exceptions:
            assert origin == swhid_origin, (
                f"the origin we guessed from the deposit db or revision ({origin}) "
                f"doesn't match the one in the deposit db's SWHID ({swhid})"
            )

    authority = MetadataAuthority(
        type=MetadataAuthorityType.DEPOSIT_CLIENT, url=provider_url, metadata={},
    )

    for (date, format, metadata) in metadata_entries:
        load_metadata(
            storage,
            row["id"],
            row["directory"],
            date,
            metadata,
            format,
            authority=authority,
            origin=origin,
            dry_run=dry_run,
        )

    return (origin, discovery_date)


def handle_row(row: Dict[str, Any], storage, deposit_cur, dry_run: bool):
    type_ = row["type"]

    # default date in case we can't find a better one
    discovery_date = row["date"] or row["committer_date"]

    metadata = row["metadata"]

    if metadata is None:
        return

    if type_ == "dsc":
        origin = None  # it will be defined later, using debian_origins_from_row

        # TODO: the debian loader writes the changelog date as the revision's
        # author date and committer date. Instead, we should use the visit's date

        if "extrinsic" in metadata:
            extrinsic_files = metadata["extrinsic"]["raw"]["files"]
            for artifact_entry in metadata["original_artifact"]:
                extrinsic_file = extrinsic_files[artifact_entry["filename"]]
                for key in ("sha256",):
                    assert artifact_entry["checksums"][key] == extrinsic_file[key]
                    artifact_entry["url"] = extrinsic_file["uri"]
            del metadata["extrinsic"]

    elif type_ == "tar":
        provider = metadata.get("extrinsic", {}).get("provider")
        if provider is not None:
            # This is the format all the package loaders currently write, and
            # it is the easiest, thanks to the 'provider' and 'when' fields,
            # which have all the information we need to tell them easily
            # and generate accurate metadata

            discovery_date = iso8601.parse_date(metadata["extrinsic"]["when"])

            # New versions of the loaders write the provider; use it.
            if provider.startswith("https://replicate.npmjs.com/"):
                # npm loader format 1

                parsed_url = urlparse(provider)
                assert re.match("^/[^/]+/?$", parsed_url.path), parsed_url
                package_name = unquote(parsed_url.path.strip("/"))
                origin = "https://www.npmjs.com/package/" + package_name
                assert_origin_exists(storage, origin)

                load_metadata(
                    storage,
                    row["id"],
                    row["directory"],
                    discovery_date,
                    metadata["extrinsic"]["raw"],
                    NPM_FORMAT,
                    authority=AUTHORITIES["npmjs"],
                    origin=origin,
                    dry_run=dry_run,
                )
                del metadata["extrinsic"]

            elif provider.startswith("https://pypi.org/"):
                # pypi loader format 1

                match = re.match(
                    "https://pypi.org/pypi/(?P<project_name>.*)/json", provider
                )
                assert match, f"unexpected provider URL format: {provider}"
                project_name = match.group("project_name")
                origin = f"https://pypi.org/project/{project_name}/"
                assert_origin_exists(storage, origin)

                load_metadata(
                    storage,
                    row["id"],
                    row["directory"],
                    discovery_date,
                    metadata["extrinsic"]["raw"],
                    PYPI_FORMAT,
                    authority=AUTHORITIES["pypi"],
                    origin=origin,
                    dry_run=dry_run,
                )
                del metadata["extrinsic"]

            elif provider.startswith("https://cran.r-project.org/"):
                # cran loader

                provider = metadata["extrinsic"]["provider"]
                if provider.startswith("https://cran.r-project.org/package="):
                    origin = metadata["extrinsic"]["provider"]
                else:
                    package_name = cran_package_from_url(provider)
                    origin = f"https://cran.r-project.org/package={package_name}"
                assert origin is not None

                # Ideally we should assert the origin exists, but we can't:
                # https://forge.softwareheritage.org/T2536
                if (
                    hashlib.sha1(origin.encode()).digest() not in _origins
                    and storage.origin_get([origin])[0] is None
                ):
                    return

                raw_extrinsic_metadata = metadata["extrinsic"]["raw"]

                # this is actually intrinsic, ignore it
                if "version" in raw_extrinsic_metadata:
                    del raw_extrinsic_metadata["version"]

                # Copy the URL to the original_artifacts metadata
                assert len(metadata["original_artifact"]) == 1
                if "url" in metadata["original_artifact"][0]:
                    assert (
                        metadata["original_artifact"][0]["url"]
                        == raw_extrinsic_metadata["url"]
                    ), row
                else:
                    metadata["original_artifact"][0]["url"] = raw_extrinsic_metadata[
                        "url"
                    ]
                del raw_extrinsic_metadata["url"]

                assert (
                    raw_extrinsic_metadata == {}
                ), f"Unexpected metadata keys: {list(raw_extrinsic_metadata)}"

                del metadata["extrinsic"]

            elif (
                provider.startswith("https://nix-community.github.io/nixpkgs-swh/")
                or provider == "https://guix.gnu.org/sources.json"
            ):
                # nixguix loader
                origin = provider
                assert_origin_exists(storage, origin)

                authority = MetadataAuthority(
                    type=MetadataAuthorityType.FORGE, url=provider, metadata={},
                )
                assert row["date"] is None  # the nixguix loader does not write dates

                load_metadata(
                    storage,
                    row["id"],
                    row["directory"],
                    discovery_date,
                    metadata["extrinsic"]["raw"],
                    NIXGUIX_FORMAT,
                    authority=authority,
                    origin=origin,
                    dry_run=dry_run,
                )
                del metadata["extrinsic"]

            elif provider.startswith("https://ftp.gnu.org/"):
                # archive loader format 1

                origin = provider
                assert_origin_exists(storage, origin)

                assert len(metadata["original_artifact"]) == 1
                metadata["original_artifact"][0]["url"] = metadata["extrinsic"]["raw"][
                    "url"
                ]

                # Remove duplicate keys of original_artifacts
                for key in ("url", "time", "length", "version", "filename"):
                    del metadata["extrinsic"]["raw"][key]

                assert metadata["extrinsic"]["raw"] == {}
                del metadata["extrinsic"]

            elif provider.startswith("https://deposit.softwareheritage.org/"):
                origin = metadata["extrinsic"]["raw"]["origin"]["url"]
                assert_origin_exists(storage, origin)

                if "@xmlns" in metadata:
                    assert metadata["@xmlns"] == ATOM_NS
                    assert metadata["@xmlns:codemeta"] in (CODEMETA_NS, [CODEMETA_NS])
                    assert "intrinsic" not in metadata
                    assert "extra_headers" not in metadata

                    # deposit loader format 1
                    # in this case, the metadata seems to be both directly in metadata
                    # and in metadata["extrinsic"]["raw"]["metadata"]

                    (origin, discovery_date) = handle_deposit_row(
                        row, discovery_date, origin, storage, deposit_cur, dry_run
                    )

                    remove_atom_codemeta_metadata_with_xmlns(metadata)
                    if "client" in metadata:
                        del metadata["client"]
                    del metadata["extrinsic"]
                else:
                    # deposit loader format 2
                    actual_metadata = metadata["extrinsic"]["raw"]["origin_metadata"][
                        "metadata"
                    ]
                    if isinstance(actual_metadata, str):
                        # new format introduced in
                        # https://forge.softwareheritage.org/D4105
                        actual_metadata = json.loads(actual_metadata)
                    if "@xmlns" in actual_metadata:
                        assert actual_metadata["@xmlns"] == ATOM_NS
                        assert actual_metadata["@xmlns:codemeta"] in (
                            CODEMETA_NS,
                            [CODEMETA_NS],
                        )
                    elif "{http://www.w3.org/2005/Atom}id" in actual_metadata:
                        assert (
                            "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}author"
                            in actual_metadata
                        )
                    else:
                        # new format introduced in
                        # https://forge.softwareheritage.org/D4065
                        # it's the same as the first case, but with the @xmlns
                        # declarations stripped
                        # Most of them should have the "id", but some revisions,
                        # like 4d3890004fade1f4ec3bf7004a4af0c490605128, are missing
                        # this field
                        assert (
                            "id" in actual_metadata
                            or "title" in actual_metadata
                            or "atom:title" in actual_metadata
                        )
                        assert "codemeta:author" in actual_metadata

                    (origin, discovery_date) = handle_deposit_row(
                        row, discovery_date, origin, storage, deposit_cur, dry_run
                    )

                    del metadata["extrinsic"]
            else:
                assert False, f"unknown provider {provider}"

        # Older versions don't write the provider; use heuristics instead.
        elif (
            metadata.get("package_source", {})
            .get("url", "")
            .startswith("https://registry.npmjs.org/")
        ):
            # npm loader format 2

            package_source_url = metadata["package_source"]["url"]
            package_name = npm_package_from_source_url(package_source_url)
            origin = "https://www.npmjs.com/package/" + package_name
            assert_origin_exists(storage, origin)

            load_metadata(
                storage,
                row["id"],
                row["directory"],
                discovery_date,
                metadata["package"],
                NPM_FORMAT,
                authority=AUTHORITIES["npmjs"],
                origin=origin,
                dry_run=dry_run,
            )
            del metadata["package"]

            assert "original_artifact" not in metadata

            # rebuild an "original_artifact"-like metadata dict from what we
            # can salvage of "package_source"
            package_source_metadata = metadata["package_source"]
            keep_keys = {"blake2s256", "filename", "sha1", "sha256", "url"}
            discard_keys = {
                "date",  # is equal to the revision date
                "name",  # was loaded above
                "version",  # same
            }
            assert (
                set(package_source_metadata) == keep_keys | discard_keys
            ), package_source_metadata

            # will be loaded below
            metadata["original_artifact"] = [
                {
                    "filename": package_source_metadata["filename"],
                    "checksums": {
                        "sha1": package_source_metadata["sha1"],
                        "sha256": package_source_metadata["sha256"],
                        "blake2s256": package_source_metadata["blake2s256"],
                    },
                    "url": package_source_metadata["url"],
                }
            ]
            del metadata["package_source"]

        elif "@xmlns" in metadata:
            assert metadata["@xmlns:codemeta"] in (CODEMETA_NS, [CODEMETA_NS])
            assert "intrinsic" not in metadata
            assert "extra_headers" not in metadata

            # deposit loader format 3

            if row["message"] == b"swh: Deposit 159 in collection swh":
                # There is no deposit 159 in the deposit DB, for some reason
                assert (
                    hash_to_hex(row["id"]) == "8e9cee14a6ad39bca4347077b87fb5bbd8953bb1"
                )
                return
            elif row["message"] == b"hal: Deposit 342 in collection hal":
                # They have status 'failed' and no swhid
                return

            origin = None  # TODO
            discovery_date = None  # TODO

            (origin, discovery_date) = handle_deposit_row(
                row, discovery_date, origin, storage, deposit_cur, dry_run
            )
            remove_atom_codemeta_metadata_with_xmlns(metadata)
            if "client" in metadata:
                del metadata["client"]  # found in the deposit db
            if "committer" in metadata:
                del metadata["committer"]  # found on the revision object

        elif "{http://www.w3.org/2005/Atom}id" in metadata:
            assert (
                "{https://doi.org/10.5063/SCHEMA/CODEMETA-2.0}author" in metadata
                or "{http://www.w3.org/2005/Atom}author" in metadata
            )
            assert "intrinsic" not in metadata
            assert "extra_headers" not in metadata

            # deposit loader format 4

            origin = None
            discovery_date = None  # TODO

            (origin, discovery_date) = handle_deposit_row(
                row, discovery_date, origin, storage, deposit_cur, dry_run
            )
            remove_atom_codemeta_metadata_without_xmlns(metadata)

        elif hash_to_hex(row["id"]) == "a86747d201ab8f8657d145df4376676d5e47cf9f":
            # deposit 91, is missing "{http://www.w3.org/2005/Atom}id" for some
            # reason, and has an invalid oririn
            return

        elif (
            isinstance(metadata.get("original_artifact"), dict)
            and metadata["original_artifact"]["url"].startswith(
                "https://files.pythonhosted.org/"
            )
        ) or (
            isinstance(metadata.get("original_artifact"), list)
            and len(metadata.get("original_artifact")) == 1
            and metadata["original_artifact"][0]
            .get("url", "")
            .startswith("https://files.pythonhosted.org/")
        ):
            if isinstance(metadata.get("original_artifact"), dict):
                metadata["original_artifact"] = [metadata["original_artifact"]]

            assert len(metadata["original_artifact"]) == 1

            origin = pypi_origin_from_filename(
                storage, row["id"], metadata["original_artifact"][0]["filename"]
            )

            if "project" in metadata:
                # pypi loader format 2
                load_metadata(
                    storage,
                    row["id"],
                    row["directory"],
                    discovery_date,
                    metadata["project"],
                    PYPI_FORMAT,
                    authority=AUTHORITIES["pypi"],
                    origin=origin,
                    dry_run=dry_run,
                )
                del metadata["project"]
            else:
                assert set(metadata) == {"original_artifact"}, set(metadata)
                # pypi loader format 3
                pass  # nothing to do, there's no metadata

        elif row["message"] == b"synthetic revision message":
            assert isinstance(metadata["original_artifact"], list), metadata
            assert not any("url" in d for d in metadata["original_artifact"])

            # archive loader format 2

            origin = None

        elif deposit_revision_message_re.match(row["message"]):
            # deposit without metadata in the revision

            assert set(metadata) == {"original_artifact"}, metadata

            origin = None  # TODO
            discovery_date = None

            (origin, discovery_date) = handle_deposit_row(
                row, discovery_date, origin, storage, deposit_cur, dry_run
            )
        else:
            assert False, f"Unable to detect type of metadata for row: {row}"

    # Ignore common intrinsic metadata keys
    for key in ("intrinsic", "extra_headers"):
        if key in metadata:
            del metadata[key]

    # Ignore loader-specific intrinsic metadata keys
    if type_ == "hg":
        del metadata["node"]
    elif type_ == "dsc":
        if "package_info" in metadata:
            del metadata["package_info"]

    if "original_artifact" in metadata:
        for original_artifact in metadata["original_artifact"]:
            # Rename keys to the expected format of original-artifacts-json.
            rename_keys = [
                ("name", "filename"),  # eg. from old Debian loader
                ("size", "length"),  # eg. from old PyPI loader
            ]
            for (old_name, new_name) in rename_keys:
                if old_name in original_artifact:
                    assert new_name not in original_artifact
                    original_artifact[new_name] = original_artifact.pop(old_name)

            # Move the checksums to their own subdict, which is the expected format
            # of original-artifacts-json.
            if "sha1" in original_artifact:
                assert "checksums" not in original_artifact
                original_artifact["checksums"] = {}
                for key in ("sha1", "sha256", "sha1_git", "blake2s256"):
                    if key in original_artifact:
                        original_artifact["checksums"][key] = original_artifact.pop(key)

            if "date" in original_artifact:
                # The information comes from the package repository rather than SWH,
                # so it shouldn't be in the 'original-artifacts' metadata
                # (which has SWH as authority).
                # Moreover, it's not a very useful information, so let's just drop it.
                del original_artifact["date"]

            allowed_keys = {
                "checksums",
                "filename",
                "length",
                "url",
                "archive_type",
            }
            assert set(original_artifact) <= allowed_keys, set(original_artifact)

        if type_ == "dsc":
            assert origin is None
            origins = debian_origins_from_row(row, storage)
            if not origins:
                print("Missing Debian origin for revision: {hash_to_hex(row['id'])}")
        else:
            origins = [origin]

        for origin in origins:
            load_metadata(
                storage,
                row["id"],
                row["directory"],
                discovery_date,
                metadata["original_artifact"],
                ORIGINAL_ARTIFACT_FORMAT,
                authority=AUTHORITIES["swh"],
                origin=origin,
                dry_run=dry_run,
            )
        del metadata["original_artifact"]

    assert metadata == {}, (
        f"remaining metadata keys for {row['id'].hex()} (type: {row['type']}): "
        f"{metadata}"
    )


def create_fetchers(db):
    with db.cursor() as cur:
        cur.execute(
            """
            INSERT INTO metadata_fetcher (name, version, metadata)
            VALUES (%s, %s, %s)
            ON CONFLICT DO NOTHING
            """,
            (FETCHER.name, FETCHER.version, FETCHER.metadata),
        )


def iter_revision_rows(storage_dbconn: str, first_id: Sha1Git):
    after_id = first_id
    failures = 0
    while True:
        try:
            storage_db = BaseDb.connect(storage_dbconn)
            with storage_db.cursor() as cur:
                while True:
                    cur.execute(
                        f"SELECT {', '.join(REVISION_COLS)} FROM revision "
                        f"WHERE id > %s AND metadata IS NOT NULL AND type != 'git'"
                        f"ORDER BY id LIMIT 1000",
                        (after_id,),
                    )
                    new_rows = 0
                    for row in cur:
                        new_rows += 1
                        row_d = dict(zip(REVISION_COLS, row))
                        yield row_d
                    after_id = row_d["id"]
                    if new_rows == 0:
                        return
        except psycopg2.OperationalError as e:
            print(e)
            # most likely a temporary error, try again
            if failures >= 60:
                raise
            else:
                time.sleep(60)
                failures += 1


def main(storage_dbconn, storage_url, deposit_dbconn, first_id, dry_run):
    storage_db = BaseDb.connect(storage_dbconn)
    deposit_db = BaseDb.connect(deposit_dbconn)
    storage = get_storage(
        "pipeline",
        steps=[
            {"cls": "retry"},
            {
                "cls": "postgresql",
                "db": storage_dbconn,
                "objstorage": {"cls": "memory", "args": {}},
            },
        ],
    )

    if not dry_run:
        create_fetchers(storage_db)
        # Not creating authorities, as the loaders are presumably already running
        # and created them already.
        # This also helps make sure this script doesn't accidentally create
        # authorities that differ from what the loaders use.

    total_rows = 0
    with deposit_db.cursor() as deposit_cur:
        for row in iter_revision_rows(storage_dbconn, first_id):
            handle_row(row, storage, deposit_cur, dry_run)

            total_rows += 1

            if total_rows % 1000 == 0:
                percents = (
                    int.from_bytes(row["id"][0:4], byteorder="big") * 100 / (1 << 32)
                )
                print(
                    f"Processed {total_rows/1000000.:.2f}M rows "
                    f"(~{percents:.1f}%, last revision: {row['id'].hex()})"
                )


if __name__ == "__main__":
    if len(sys.argv) == 4:
        (_, storage_dbconn, storage_url, deposit_dbconn) = sys.argv
        first_id = "00" * 20
    elif len(sys.argv) == 5:
        (_, storage_dbconn, storage_url, deposit_dbconn, first_id) = sys.argv
    else:
        print(
            f"Syntax: {sys.argv[0]} <storage_dbconn> <storage_url> "
            f"<deposit_dbconn> [<first id>]"
        )
        exit(1)

    if os.path.isfile("./origins.txt"):
        # You can generate this file with:
        # psql service=swh-replica \
        #   -c "\copy (select digest(url, 'sha1') from origin) to stdout" \
        #   | pv -l > origins.txt
        print("Loading origins...")
        with open("./origins.txt") as fd:
            for line in fd:
                digest = line.strip()[3:]
                _origins.add(bytes.fromhex(digest))
        print("Done loading origins.")

    main(storage_dbconn, storage_url, deposit_dbconn, bytes.fromhex(first_id), True)
